

class HookManager:
    def __init__(self):
        self._original_callables = {}

    def attach_hooks_on_module(self, module, predicate, hook_creator):
        self.attach_hooks_on_module_using(
            module, module, predicate, hook_creator)

    def attach_hooks_on_module_using(
            self, module, using_module, predicate, hook_creator):
        """
        Attach hooks onto functions in the provided module. Use the
        `using_module` to discover the existing functions.
        """
        for prop in dir(using_module):
            if not predicate(getattr(module, prop)):
                continue
            self.attach_hook(module, prop, hook_creator)

    def attach_hook(self, module, prop, hook_creator):
        target = getattr(module, prop)
        self._maybe_store_callable(module, prop, target)
        setattr(module, prop, hook_creator(target))

    def remove_hooks(self):
        for module, callable_pairs in self._original_callables.items():
            for prop, original_callable in callable_pairs.items():
                setattr(module, prop, original_callable)
        self._original_callables.clear()

    def _maybe_store_callable(self, module, prop, original_callable):
        """
        Store the original callable (to be able to restore it) only when it is
        the first time we are encountering the given callable.
        """
        if module not in self._original_callables:
            self._original_callables[module] = {}

        if prop in self._original_callables[module]:
            return

        self._original_callables[module][prop] = original_callable



import contextlib
import sqlite3


class TrackerBase:
    def __init__(self):
        self._is_tracking = False

    @contextlib.contextmanager
    def track(self):
        self.start_tracking()
        try:
            yield self
        finally:
            self.stop_tracking()

    def start_tracking(self):
        self._is_tracking = True

    def stop_tracking(self):
        self._is_tracking = False

    def populate_report(self, builder):
        raise NotImplementedError


class ReportBase:
    def __init__(self, connection):
        self._connection = connection

    def __del__(self):
        self._connection.close()


class ReportBuilderBase:
    def __init__(self, file=None):
        database_file = file if file is not None else ':memory:'
        self._connection = sqlite3.connect(database_file, check_same_thread=False)
        self._create_report_tables()

    def process_tracker(self, tracker):
        tracker.populate_report(self)
        return self

    def build(self):
        raise NotImplementedError

    def _create_report_tables(self):
        raise NotImplementedError



import collections
import enum

from deepview_profile.tracking.base import ReportBase, ReportBuilderBase
import deepview_profile.tracking.memory.report_queries as queries


WeightEntry = collections.namedtuple(
    'WeightEntry',
    ['weight_name',
     'size_bytes',
     'grad_size_bytes',
     'file_path',
     'line_number'],
)


ActivationEntry = collections.namedtuple(
    'ActivationEntry',
    ['operation_name', 'size_bytes', 'file_path', 'line_number'],
)


class MiscSizeType(enum.Enum):
    PeakUsageBytes = 'peak_usage_bytes'


class MemoryReport(ReportBase):
    def __init__(self, connection):
        super().__init__(connection)

    def get_weight_entries(self, path_prefix=None):
        cursor = self._connection.cursor()
        return map(
            lambda row: WeightEntry(*row),
            cursor.execute(queries.get_weight_entries_with_context),
        )

    def get_activation_entries(self, path_prefix=None):
        cursor = self._connection.cursor()
        return map(
            lambda row: ActivationEntry(*row),
            cursor.execute(queries.get_activation_entries_with_context),
        )

    def get_misc_entry(self, misc_size_type: MiscSizeType):
        cursor = self._connection.cursor()
        cursor.execute(queries.get_misc_entry, (misc_size_type.value,))
        return cursor.fetchone()[0]


class MemoryReportBuilder(ReportBuilderBase):
    # This is the memory tracking report file format version that will be
    # created by this builder. When changes are made to the file format, this
    # integer should be increased monotonically.
    #
    # We need to version these tracking reports to protect us from future
    # changes to the file format.
    Version = 1

    def __init__(self, file=None):
        super().__init__(file)

    def add_weight_entry(
            self, weight_name, size_bytes, grad_size_bytes, stack_context):
        cursor = self._connection.cursor()
        cursor.execute(
            queries.add_weight_entry,
            (weight_name, size_bytes, grad_size_bytes),
        )
        self._add_stack_frames(
            cursor=cursor,
            entry_id=cursor.lastrowid,
            entry_type=queries.EntryType.Weight,
            stack_context=stack_context,
        )
        return self

    def add_activation_entry(self, operation_name, size_bytes, stack_context):
        cursor = self._connection.cursor()
        cursor.execute(
            queries.add_activation_entry, (operation_name, size_bytes))
        self._add_stack_frames(
            cursor=cursor,
            entry_id=cursor.lastrowid,
            entry_type=queries.EntryType.Activation,
            stack_context=stack_context,
        )
        return self

    def add_misc_entry(self, size_type: MiscSizeType, size_bytes):
        cursor = self._connection.cursor()
        cursor.execute(queries.add_misc_entry, (size_type.value, size_bytes))
        return self

    def build(self):
        self._connection.commit()
        return MemoryReport(self._connection)

    def _create_report_tables(self):
        cursor = self._connection.cursor()
        cursor.execute(queries.set_report_format_version.format(
            version=MemoryReportBuilder.Version))
        for creation_query in queries.create_report_tables.values():
            cursor.execute(creation_query)
        cursor.executemany(
            queries.add_entry_type,
            map(lambda entry: (entry.value, entry.name), queries.EntryType),
        )
        self._connection.commit()

    def _add_stack_frames(
        self,
        cursor,
        entry_id,
        entry_type: queries.EntryType,
        stack_context,
    ):
        cursor.execute(
            queries.add_correlation_entry, (entry_id, entry_type.value))
        correlation_id = cursor.lastrowid

        def stack_frame_generator():
            for idx, frame in enumerate(stack_context.frames):
                yield (correlation_id, idx, frame.file_path, frame.line_number)

        cursor.executemany(queries.add_stack_frame, stack_frame_generator())



import collections
import logging
import torch

from deepview_profile.exceptions import AnalysisError
from deepview_profile.user_code_utils import user_code_environment
from deepview_profile.utils import release_memory

logger = logging.getLogger(__name__)

IterationSample = collections.namedtuple(
    "IterationSample", ["batch_size", "run_time_ms", "peak_usage_bytes"])


class IterationProfiler:
    def __init__(
        self,
        iteration,
        input_provider,
        path_to_entry_point_dir,
        project_root
    ):
        self._iteration = iteration
        self._input_provider = input_provider
        self._path_to_entry_point_dir = path_to_entry_point_dir
        self._project_root = project_root
        self._start_event = torch.cuda.Event(enable_timing=True)
        self._end_event = torch.cuda.Event(enable_timing=True)

    @classmethod
    def new_from(
        cls,
        model_provider,
        input_provider,
        iteration_provider,
        path_to_entry_point_dir,
        project_root,
    ):
        with user_code_environment(path_to_entry_point_dir, project_root):
            model = model_provider()
            iteration = iteration_provider(model)
        return cls(
            iteration, input_provider, path_to_entry_point_dir, project_root)

    def measure_run_time_ms(self, batch_size, initial_repetitions=None):
        """
        Measures the iteration run time in milliseconds.

        NOTE: This method will raise a RuntimeError if there is not enough GPU
              memory to run the iteration.
        """

        with user_code_environment(
                self._path_to_entry_point_dir, self._project_root):
            inputs = self._input_provider(batch_size=batch_size)
            # Warm up - 10 iterations
            for _ in range(10):
                self._iteration(*inputs)

        torch.cuda.synchronize()

        def measure(iterations):
            with user_code_environment(
                    self._path_to_entry_point_dir, self._project_root):
                self._start_event.record()
                for _ in range(iterations):
                    self._iteration(*inputs)
                self._end_event.record()
            torch.cuda.synchronize()
            return self._start_event.elapsed_time(self._end_event)

        # When measuring the iteration run time of the model, we want to use as
        # few repetitions as possible. The problem is that we do not know how
        # many repetitions we need ahead of time.
        #
        # So the idea here is to iteratively double the number of repetitions
        # we use until we get a stable measurement (less than 5% difference).
        # If the caller knows how many measurements to use, they can pass it in
        # and we'll start from there. We will stop after reaching 100
        # iterations (or 2 * initial_repetitions), whichever is larger.
        #
        # We will return the smaller number of repetitions. This can be used by
        # future callers as the value of the initial_repetitions argument.
        repetitions = 3 if initial_repetitions is None else initial_repetitions
        max_repetitions = (
            50 if initial_repetitions is None else max(50, initial_repetitions)
        )

        torch.cuda.reset_max_memory_allocated()
        lesser = measure(repetitions) / repetitions
        peak_usage_bytes = torch.cuda.max_memory_allocated()

        # in the worst case, run for at least 2 seconds before giving up
        min_profile_time_ms = 2000
        max_repetitions = max(max_repetitions, min_profile_time_ms / lesser)

        logger.debug("Iters: %d, Measured: %f", repetitions, lesser)
        while repetitions <= max_repetitions:
            doubled = repetitions * 2
            greater = measure(doubled) / doubled
            logger.debug("Iters: %d, Measured: %f (range: %f)",\
                         doubled, greater, max(lesser, greater) / min(lesser, greater))

            # Stop when the difference between the measurements is less than 5%
            if (max(lesser, greater) / min(lesser, greater)) < 1.05:
                break

            repetitions = doubled
            lesser = greater

        return min(lesser, greater), peak_usage_bytes, repetitions

    def measure_run_time_ms_catch_oom(
            self, batch_size, initial_repetitions=None):
        # This function is useful when we want to explicitly handle OOM errors
        # without aborting the profiling.
        release_memory()
        try:
            return (
                None,
                self.measure_run_time_ms(batch_size, initial_repetitions),
            )
        except AnalysisError as ex:
            message = str(ex)
            if 'CUDA out of memory' in message:
                return (ex, None)
            else:
                raise

    def sample_run_time_ms_by_batch_size(
        self,
        start_batch_size,
        start_batch_size_run_time_ms=None,
        start_batch_size_peak_usage_bytes=None,
        memory_usage_percentage=None,
        num_samples=3,
    ):
        samples = []

        # 1. Make sure we can measure the run time of the "start" batch size
        if (start_batch_size_run_time_ms is None or
                start_batch_size_peak_usage_bytes is None):
            start_run_time_ms, start_peak_usage_bytes, _ = \
                self.measure_run_time_ms(start_batch_size)
        else:
            start_run_time_ms = start_batch_size_run_time_ms
            start_peak_usage_bytes = start_batch_size_peak_usage_bytes

        samples.append(IterationSample(
            start_batch_size, start_run_time_ms, start_peak_usage_bytes))

        # 2. Perform sampling. We keep a range of "viable" batch sizes, where
        #    the upper limit is a guess on what will fit in memory. We adjust
        #    these limits as we sample.
        #
        #    We estimate the maximum batch size by assuming a linear
        #    relationship between the model's memory use and its batch size.
        #    The memory use at the initial batch size is obtained through
        #    memory tracking. If it is not specified, we just add a constant
        max_batch_size = (
            start_batch_size / memory_usage_percentage
            if memory_usage_percentage is not None
            else start_batch_size + 100
        )

        if len(samples) < num_samples:
            samples.extend(self._sample_range(
                start_batch_size,
                max_batch_size,
                num_samples=(num_samples - len(samples)),
                is_increasing=True,
            ))

        if len(samples) < num_samples:
            samples.extend(self._sample_range(
                1,
                start_batch_size,
                num_samples=(num_samples - len(samples)),
                is_increasing=False,
            ))

        return samples

    def _sample_range(
            self, min_size, max_size, num_samples, is_increasing=True):
        # The idea here is to sample the range of possible batch sizes by
        # recursively narrowing down the acceptable ranges of batch sizes.

        samples = []
        stack = [(min_size, max_size)]

        while len(samples) < num_samples and len(stack) > 0:
            lower, upper = stack.pop()
            if lower >= upper:
                continue

            next_size = self._select_batch_size(lower, upper, is_increasing)
            logger.debug(
                "[%d, %d] Sampling batch size: %d", lower, upper, next_size)
            err, result = self.measure_run_time_ms_catch_oom(next_size)
            if err is not None:
                stack.append((lower, next_size - 1))
                continue

            samples.append(IterationSample(next_size, result[0], result[1]))

            # Change the order in which we explore each range
            if is_increasing:
                stack.append((lower, next_size - 1))
                stack.append((next_size + 1, upper))
            else:
                stack.append((next_size + 1, upper))
                stack.append((lower, next_size - 1))

        return samples

    def _select_batch_size(self, lower, upper, is_increasing):
        diff = upper - lower
        base = lower if is_increasing else upper
        mult = 1 if is_increasing else -1

        # increase the growth amount for batch sizes to better account for small
        # models whose initial batch size is much smaller than the maximum possible.
        tiers = [100, 20, 10, 5]
        for t in tiers:
            if diff >= t:
                return base + mult * t
        return base + mult


import os


class Sentinel:
    def __init__(self):
        self._read_pipe = None
        self._write_pipe = None

    def start(self):
        self._read_pipe, self._write_pipe = os.pipe()

    def stop(self):
        os.close(self._write_pipe)
        os.close(self._read_pipe)
        self._read_pipe = None
        self._write_pipe = None

    @property
    def read_pipe(self):
        return self._read_pipe

    def should_exit(self, ready_descriptors):
        return self._read_pipe in ready_descriptors

    def signal_exit(self):
        os.write(self._write_pipe, b'\0')

    def consume_exit_signal(self):
        # This should only be called after signal_exit(),
        # otherwise the calling thread will block.
        os.read(self._read_pipe, 1)



import ast
import inspect
import logging
import math
import os
import pynvml
import re
import torch
import numpy as np
import deepview_profile.protocol_gen.innpv_pb2 as pm
from deepview_profile.analysis.static import StaticAnalyzer
from deepview_profile.db.database import DatabaseInterface, EnergyTableInterface
from deepview_profile.energy.measurer import EnergyMeasurer
from deepview_profile.exceptions import AnalysisError, exceptions_as_analysis_errors
from deepview_profile.profiler.iteration import IterationProfiler
from deepview_profile.tracking.tracker import Tracker
from deepview_profile.user_code_utils import user_code_environment
from deepview_profile.profiler.utilization import (
    utilization_analysis,
    serialize_response,
)
from deepview_profile.profiler.ddp import ddp_analysis
from typing import List

logger = logging.getLogger(__name__)

# habitat imports
try:
    import collections
    import habitat
    import habitat.habitat_cuda as hc
    from habitat.analysis import SPECIAL_OPERATIONS
    from habitat.profiling.run_time import RunTimeProfiler

    habitat_found = True
except ImportError:
    logger.debug("DeepView.Predict not found, GPU predictions not available")
    habitat_found = False


MODEL_PROVIDER_NAME = "deepview_model_provider"
INPUT_PROVIDER_NAME = "deepview_input_provider"
ITERATION_PROVIDER_NAME = "deepview_iteration_provider"
BATCH_SIZE_ARG = "batch_size"


# Habitat variables
Context = collections.namedtuple(
    "Context",
    ["origin_device", "profiler", "percentile"],
)


class AnalysisSession:
    def __init__(
        self,
        project_root,
        entry_point,
        path_to_entry_point_dir,
        model_provider,
        input_provider,
        iteration_provider,
        batch_size,
        entry_point_static_analyzer,
    ):
        self._project_root = project_root
        self._entry_point = entry_point
        self._path_to_entry_point_dir = path_to_entry_point_dir
        self._model_provider = model_provider
        self._input_provider = input_provider
        self._iteration_provider = iteration_provider
        self._batch_size = batch_size
        self._entry_point_static_analyzer = entry_point_static_analyzer
        self._profiler = None
        self._memory_usage_percentage = None
        self._batch_size_iteration_run_time_ms = None
        self._batch_size_peak_usage_bytes = None
        self._utilization = None
        self._energy_table_interface = EnergyTableInterface(
            DatabaseInterface().connection
        )

    @classmethod
    def new_from(cls, project_root, entry_point):
        path_to_entry_point = os.path.join(project_root, entry_point)
        # Note: This is not necessarily the same as project_root because the
        #       entry_point could be in a subdirectory.
        path_to_entry_point_dir = os.path.dirname(path_to_entry_point)

        # 1. Run the entry point file to "load" the model
        entry_point_code, entry_point_ast, scope = _run_entry_point(
            path_to_entry_point,
            path_to_entry_point_dir,
            project_root,
        )

        # 2. Check that the model provider and input provider functions exist
        if MODEL_PROVIDER_NAME not in scope:
            raise AnalysisError(
                "The project entry point file is missing a model provider "
                "function. Please add a model provider function named "
                '"{}".'.format(MODEL_PROVIDER_NAME)
            ).with_file_context(entry_point)

        if INPUT_PROVIDER_NAME not in scope:
            raise AnalysisError(
                "The project entry point file is missing an input provider "
                "function. Please add an input provider function named "
                '"{}".'.format(INPUT_PROVIDER_NAME)
            ).with_file_context(entry_point)

        if ITERATION_PROVIDER_NAME not in scope:
            raise AnalysisError(
                "The project entry point file is missing an iteration "
                "provider function. Please add an iteration provider function "
                'named "{}".'.format(ITERATION_PROVIDER_NAME)
            ).with_file_context(entry_point)

        batch_size = _validate_providers_signatures(
            scope[MODEL_PROVIDER_NAME],
            scope[INPUT_PROVIDER_NAME],
            scope[ITERATION_PROVIDER_NAME],
            entry_point,
        )

        (
            model_provider,
            input_provider,
            iteration_provider,
        ) = _wrap_providers_with_validators(
            scope[MODEL_PROVIDER_NAME],
            scope[INPUT_PROVIDER_NAME],
            scope[ITERATION_PROVIDER_NAME],
            entry_point,
        )

        return cls(
            project_root,
            entry_point,
            path_to_entry_point_dir,
            model_provider,
            input_provider,
            iteration_provider,
            batch_size,
            StaticAnalyzer(entry_point_code, entry_point_ast),
        )

    def measure_utilization(self):
        resp = pm.UtilizationResponse()
        try:
            analysis_results = utilization_analysis(
                self._model_provider, self._input_provider, self._iteration_provider
            )
            serialize_response(resp.rootNode, analysis_results["root_node"])
            resp.tensor_utilization = float(analysis_results["tensor_core_perc"])
        except AnalysisError as ex:
            message = str(ex)
            logger.error(message)
            resp.analysis_error.error_message = message
        except Exception as ex:
            message = str(ex)
            logger.error(message)
            logger.error("There was an error measuring utilization")
            resp.analysis_error.error_message = (
                "There was an error measuring utilization"
            )
        finally:
            return resp

    def ddp_computation(self):
        resp = pm.DDPBucketSizesComputationTimes()

        try:
            analysis_results = ddp_analysis(
                self._model_provider, self._input_provider, self._iteration_provider
            )
            resp.forward_time_ms = float(analysis_results["forward_time_ms"])
            resp.bucket_sizes.extend(analysis_results["bucket_sizes"])
            for computation_time_item in analysis_results["expected_computation_times"]:
                add_item_to_resp = resp.computation_times.add()
                add_item_to_resp.ngpus = int(computation_time_item["ngpus"])
                add_item_to_resp.expected_max_times.extend(
                    computation_time_item["expected_max_times"]
                )

        except AnalysisError as ex:
            message = str(ex)
            logger.error(message)
            resp.analysis_error.error_message = message
        except Exception as ex:
            message = str(ex)
            logger.error(message)
            logger.error("There was an error measuring ddp throughput")
            resp.analysis_error.error_message = (
                "There was an error measuring ddp throughput"
            )
        finally:
            return resp

    def energy_compute(self) -> pm.EnergyResponse:
        energy_measurer = EnergyMeasurer()

        model = self._model_provider()
        inputs = self._input_provider()
        iteration = self._iteration_provider(model)
        resp = pm.EnergyResponse()

        try:
            energy_measurer.begin_measurement()
            iterations = 20
            for _ in range(iterations):
                iteration(*inputs)
            energy_measurer.end_measurement()
            resp.total_consumption = energy_measurer.total_energy() / float(iterations)
            resp.batch_size = self._batch_size

            components = []
            components_joules = []

            if energy_measurer.cpu_energy() is not None:
                cpu_component = pm.EnergyConsumptionComponent()
                cpu_component.component_type = pm.ENERGY_CPU_DRAM
                cpu_component.consumption_joules = energy_measurer.cpu_energy() / float(
                    iterations
                )
                components.append(cpu_component)
                components_joules.append(cpu_component.consumption_joules)
            else:
                cpu_component = pm.EnergyConsumptionComponent()
                cpu_component.component_type = pm.ENERGY_CPU_DRAM
                cpu_component.consumption_joules = 0.0
                components.append(cpu_component)
                components_joules.append(cpu_component.consumption_joules)

            gpu_component = pm.EnergyConsumptionComponent()
            gpu_component.component_type = pm.ENERGY_NVIDIA
            gpu_component.consumption_joules = energy_measurer.gpu_energy() / float(
                iterations
            )
            components.append(gpu_component)
            components_joules.append(gpu_component.consumption_joules)

            resp.components.extend(components)

            # get last 10 runs if they exist
            path_to_entry_point = os.path.join(self._project_root, self._entry_point)
            past_runs = (
                self._energy_table_interface.get_latest_n_entries_of_entry_point(
                    10, path_to_entry_point
                )
            )
            resp.past_measurements.extend(_convert_to_energy_responses(past_runs))

            # add current run to database
            current_entry = [path_to_entry_point] + components_joules
            current_entry.append(self._batch_size)
            self._energy_table_interface.add_entry(current_entry)
        except AnalysisError as ex:
            message = str(ex)
            logger.error(message)
            resp.analysis_error.error_message = message
        except Exception as ex:
            message = str(ex)
            logger.error(message)
            resp.analysis_error.error_message = (
                f"There was an error obtaining energy measurements: {message}"
            )
        finally:
            return resp

    def habitat_compute_threshold(self, runnable, context):
        tracker = habitat.OperationTracker(context.origin_device)
        with tracker.track():
            runnable()

        run_times = []
        trace = tracker.get_tracked_trace()
        for op in trace.operations:
            if op.name in SPECIAL_OPERATIONS:
                continue
            run_times.append(op.forward.run_time_ms)
            if op.backward is not None:
                run_times.append(op.backward.run_time_ms)

        return np.percentile(run_times, context.percentile)

    def habitat_predict(self):
        resp = pm.HabitatResponse()
        if not habitat_found:
            logger.debug("Skipping deepview predictions, returning empty response.")
            return resp

        try:
            print("deepview_predict: begin")
            # clear any cupti activity before running Deepview.Predict
            hc.release_cupti_hook()
            DEVICES = [
                habitat.Device.P100,
                habitat.Device.P4000,
                habitat.Device.RTX2070,
                habitat.Device.RTX2080Ti,
                habitat.Device.T4,
                habitat.Device.V100,
                habitat.Device.A100,
                habitat.Device.RTX3090,
                habitat.Device.A40,
                habitat.Device.A4000,
                habitat.Device.RTX4000,
            ]

            # Detect source GPU
            pynvml.nvmlInit()
            if pynvml.nvmlDeviceGetCount() == 0:
                raise Exception(
                    "NVML failed to find a GPU. Please ensure that you \
                have a NVIDIA GPU installed and that the drivers are functioning \
                correctly."
                )

            # TODO: Consider profiling on not only the first detected GPU
            nvml_handle = pynvml.nvmlDeviceGetHandleByIndex(0)
            source_device_name = pynvml.nvmlDeviceGetName(nvml_handle).decode("utf-8")
            split_source_device_name = re.split(r"-|\s|_|\\|/", source_device_name)
            source_device = (
                None if logging.root.level > logging.DEBUG else habitat.Device.T4
            )
            for device in DEVICES:
                if device.name in "".join(split_source_device_name):
                    source_device = device
            pynvml.nvmlShutdown()
            if not source_device:
                logger.debug(
                    "Skipping DeepView predictions,\
                    source not in list of supported GPUs."
                )
                src = pm.HabitatDevicePrediction()
                src.device_name = "unavailable"
                src.runtime_ms = -1
                resp.predictions.append(src)
                return resp

            print("deepview_predict: detected source device", source_device.name)

            # get model
            model = self._model_provider()
            inputs = self._input_provider()
            iteration = self._iteration_provider(model)

            def runnable():
                iteration(*inputs)

            profiler = RunTimeProfiler()

            context = Context(
                origin_device=source_device, profiler=profiler, percentile=99.5
            )

            threshold = self.habitat_compute_threshold(runnable, context)

            tracker = habitat.OperationTracker(
                device=context.origin_device,
                metrics=[
                    habitat.Metric.SinglePrecisionFLOPEfficiency,
                    habitat.Metric.DRAMReadBytes,
                    habitat.Metric.DRAMWriteBytes,
                ],
                metrics_threshold_ms=threshold,
            )

            with tracker.track():
                iteration(*inputs)

            print("deepview_predict: tracing on origin device")
            trace = tracker.get_tracked_trace()

            src = pm.HabitatDevicePrediction()
            src.device_name = "source"
            src.runtime_ms = trace.run_time_ms
            resp.predictions.append(src)

            for device in DEVICES:
                print("deepview_predict: predicting for", device)
                predicted_trace = trace.to_device(device)

                pred = pm.HabitatDevicePrediction()
                pred.device_name = device.name
                pred.runtime_ms = predicted_trace.run_time_ms
                resp.predictions.append(pred)

            print(f"returning {len(resp.predictions)} predictions.")
            # clear cupti after Deepview.Predict
            hc.release_cupti_hook()
        except AnalysisError as ex:
            message = str(ex)
            logger.error(message)
            resp.analysis_error.error_message = message
        except Exception as ex:
            message = str(ex)
            logger.error("There was an error running DeepView Predict")
            logger.error(message)
            cupti_context_error = ""
            if "CUPTI_ERROR_MULTIPLE_SUBSCRIBERS_NOT_SUPPORTED" in message:
                cupti_context_error = """We have detected that cuda versions <= 11.8 maybe not close CUPTI context properly between each execution.
                Please try with a version of pytorch that includes cuda version >= 12.1. Or you may close this window and open a new one."""
                print(cupti_context_error)
            resp.analysis_error.error_message = (
                "There was an error running DeepView Predict\n" + cupti_context_error
            )
        finally:
            return resp

    def measure_breakdown(self, nvml):
        # 1. Measure the breakdown entries
        self._prepare_for_memory_profiling()
        tracker = self._get_tracker_instance()
        tracker.track_memory()
        tracker.track_run_time()
        breakdown = tracker.get_hierarchical_breakdown()
        del tracker

        # 2. Measure the overall iteration run time
        if self._batch_size_iteration_run_time_ms is None:
            if self._profiler is None:
                self._initialize_iteration_profiler()

            (
                self._batch_size_iteration_run_time_ms,
                self._batch_size_peak_usage_bytes,
                _,
            ) = self._profiler.measure_run_time_ms(self._batch_size)

        # 3. Serialize the measured data
        bm = pm.BreakdownResponse()
        bm.batch_size = self._batch_size
        bm.peak_usage_bytes = breakdown.peak_usage_bytes
        bm.memory_capacity_bytes = nvml.get_memory_capacity().total
        bm.iteration_run_time_ms = self._batch_size_iteration_run_time_ms
        breakdown.operations.serialize_to_protobuf(bm.operation_tree)
        breakdown.weights.serialize_to_protobuf(bm.weight_tree)

        # 4. Bookkeeping for the throughput measurements
        self._memory_usage_percentage = bm.peak_usage_bytes / bm.memory_capacity_bytes

        return bm

    def measure_throughput(self):
        if self._profiler is None:
            self._initialize_iteration_profiler()

        # 1. Measure the throughput at several spots to be able to build a
        #    prediction model
        num_samples = 3
        samples = self._profiler.sample_run_time_ms_by_batch_size(
            start_batch_size=self._batch_size,
            memory_usage_percentage=self._memory_usage_percentage,
            start_batch_size_run_time_ms=self._batch_size_iteration_run_time_ms,
            num_samples=num_samples,
        )
        if len(samples) == 0 or samples[0].batch_size != self._batch_size:
            raise AnalysisError(
                "Something went wrong with DeepView.Profile when measuring your "
                "model's throughput. Please file a bug."
            )

        # 2. Begin filling in the throughput response
        logger.debug("sampling results \n %r" % str(samples))
        measured_throughput = samples[0].batch_size / samples[0].run_time_ms * 1000
        throughput = pm.ThroughputResponse()
        throughput.samples_per_second = measured_throughput
        throughput.predicted_max_samples_per_second = math.nan
        throughput.can_manipulate_batch_size = False

        # 3. Determine whether we have information about the batch size
        #    location in the code
        batch_info = self._entry_point_static_analyzer.batch_size_location()
        if batch_info is not None:
            throughput.batch_size_context.line_number = batch_info[0]
            throughput.can_manipulate_batch_size = batch_info[1]
            throughput.batch_size_context.file_path.components.extend(
                self._entry_point.split(os.sep)
            )

        # 4. If we do not have enough throughput samples, we cannot build any
        #    prediction models so just return the message as-is
        if len(samples) != num_samples:
            return throughput

        # 5. Build and validate the prediction models for run time (throughput)
        #    and memory
        batches = list(map(lambda sample: sample.batch_size, samples))
        run_times = list(map(lambda sample: sample.run_time_ms, samples))
        usages = list(map(lambda sample: sample.peak_usage_bytes, samples))

        run_time_model = _fit_linear_model(batches, run_times)
        peak_usage_model = _fit_linear_model(batches, usages)

        logger.debug(
            "Run time model - Slope: %f, Bias: %f (ms)",
            *run_time_model,
        )
        logger.debug(
            "Peak usage model - Slope: %f, Bias: %f (bytes)",
            *peak_usage_model,
        )

        throughput.peak_usage_bytes.slope = peak_usage_model[0]
        throughput.peak_usage_bytes.bias = peak_usage_model[1]

        predicted_max_throughput = 1000.0 / run_time_model[0]

        # Our prediction can be inaccurate due to sampling error or incorrect
        # assumptions. In these cases, we ignore our prediction. At the very
        # minimum, a good linear model has a positive slope and bias.
        # if (run_time_model[0] < 1e-3 or run_time_model[1] < 1e-3 or
        if run_time_model[0] < 1e-3 or measured_throughput > predicted_max_throughput:
            return throughput

        throughput.predicted_max_samples_per_second = predicted_max_throughput
        throughput.run_time_ms.slope = run_time_model[0]
        throughput.run_time_ms.bias = run_time_model[1]

        return throughput

    def measure_peak_usage_bytes(self):
        self._prepare_for_memory_profiling()
        # Run one iteration to initialize the gradients
        # with user_code_environment(self._path_to_entry_point_dir, self._project_root):
        model = self._model_provider()
        iteration = self._iteration_provider(model)
        iteration(*self._input_provider(batch_size=self._batch_size))

        torch.cuda.reset_max_memory_allocated()
        # with user_code_environment(self._path_to_entry_point_dir, self._project_root):
        # NOTE: It's important to run at least 2 iterations here. It turns
        #       out that >= 2 iterations is the number of iterations needed
        #       to get a stable measurement of the total memory
        #       consumption. When using Adam, if you run one iteration, the
        #       memory usage ends up being too low by a constant factor.
        for _ in range(2):
            iteration(*(self._input_provider(batch_size=self._batch_size)))
        return torch.cuda.max_memory_allocated()

    def generate_memory_usage_report(self, save_report_to):
        self._prepare_for_memory_profiling()
        tracker = self._get_tracker_instance()
        tracker.track_memory()
        tracker.get_memory_report(report_file=save_report_to)

    def generate_run_time_breakdown_report(self, save_report_to):
        tracker = self._get_tracker_instance()
        tracker.track_run_time()
        tracker.get_run_time_report(report_file=save_report_to)

    def _initialize_iteration_profiler(self):
        self._profiler = IterationProfiler.new_from(
            self._model_provider,
            self._input_provider,
            self._iteration_provider,
            self._path_to_entry_point_dir,
            self._project_root,
        )

    def _prepare_for_memory_profiling(self):
        if self._profiler is not None:
            # It's important that the IterationProfiler is uninitialized here
            # because it stores a copy of the model, which takes up GPU memory.
            # This would skew our memory profiling results.
            del self._profiler
            self._profiler = None

    def _get_tracker_instance(self):
        return Tracker(
            model_provider=self._model_provider,
            iteration_provider=self._iteration_provider,
            input_provider=self._input_provider,
            project_root=self._project_root,
            user_code_path=self._path_to_entry_point_dir,
        )


def _run_entry_point(path_to_entry_point, path_to_entry_point_dir, project_root):
    with open(path_to_entry_point) as file:
        code_str = file.read()
    with exceptions_as_analysis_errors(project_root):
        tree = ast.parse(code_str, filename=path_to_entry_point)
        code = compile(tree, path_to_entry_point, mode="exec")
    with user_code_environment(path_to_entry_point_dir, project_root):
        scope = {}
        exec(code, scope, scope)
    return code_str, tree, scope


def _validate_providers_signatures(
    model_provider,
    input_provider,
    iteration_provider,
    entry_point,
):
    model_sig = inspect.signature(model_provider)
    if len(model_sig.parameters) != 0:
        raise AnalysisError(
            "The model provider function cannot have any parameters."
        ).with_file_context(entry_point)

    input_sig = inspect.signature(input_provider)
    if (
        len(input_sig.parameters) != 1
        or BATCH_SIZE_ARG not in input_sig.parameters
        or type(input_sig.parameters[BATCH_SIZE_ARG].default) is not int
    ):
        raise AnalysisError(
            "The input provider function must have exactly one '{}' "
            "parameter with an integral default "
            "value.".format(BATCH_SIZE_ARG)
        ).with_file_context(entry_point)
    batch_size = input_sig.parameters[BATCH_SIZE_ARG].default

    iteration_sig = inspect.signature(iteration_provider)
    if len(iteration_sig.parameters) != 1:
        raise AnalysisError(
            "The iteration provider function must have exactly one "
            "parameter (the model being profiled)."
        ).with_file_context(entry_point)

    return batch_size


def _wrap_providers_with_validators(
    model_provider,
    input_provider,
    iteration_provider,
    entry_point,
):
    def model_provider_wrapped():
        model = model_provider()
        if not callable(model):
            raise AnalysisError(
                "The model provider function must return a callable (i.e. "
                "return something that can be called like a PyTorch "
                "module or function)."
            ).with_file_context(entry_point)
        return model

    def input_provider_wrapped(batch_size=None):
        if batch_size is None:
            inputs = input_provider()
        else:
            inputs = input_provider(batch_size=batch_size)

        if isinstance(inputs, torch.Tensor):
            raise AnalysisError(
                "The input provider function must return an iterable that "
                "contains the inputs for the model. If your model only takes "
                "a single tensor as input, return a single element tuple or "
                "list in your input provider (e.g., 'return [the_input]')."
            ).with_file_context(entry_point)

        try:
            iter(inputs)
            return inputs
        except TypeError:
            raise AnalysisError(
                "The input provider function must return an iterable that "
                "contains the inputs for the model."
            ).with_file_context(entry_point)

    def iteration_provider_wrapped(model):
        iteration = iteration_provider(model)
        if not callable(iteration):
            raise AnalysisError(
                "The iteration provider function must return a callable "
                "(i.e. return something that can be called like a "
                "function)."
            ).with_file_context(entry_point)
        return iteration

    return (
        model_provider_wrapped,
        input_provider_wrapped,
        iteration_provider_wrapped,
    )


def _fit_linear_model(x, y):
    y_np = np.array(y)
    x_np = np.array(x)
    stacked = np.vstack([x_np, np.ones(len(x_np))]).T
    slope, bias = np.linalg.lstsq(stacked, y_np, rcond=None)[0]
    # Linear model: y = slope * x + bias
    return slope, bias


def _convert_to_energy_responses(entries: list) -> List[pm.EnergyResponse]:
    energy_response_list = []
    for entry in entries:
        if EnergyTableInterface.is_valid_entry_with_timestamp(entry):
            energy_response = pm.EnergyResponse()
            cpu_component = pm.EnergyConsumptionComponent()
            cpu_component.component_type = pm.ENERGY_CPU_DRAM
            cpu_component.consumption_joules = entry[1]

            gpu_component = pm.EnergyConsumptionComponent()
            gpu_component.component_type = pm.ENERGY_NVIDIA
            gpu_component.consumption_joules = entry[2]

            energy_response.total_consumption = (
                gpu_component.consumption_joules + cpu_component.consumption_joules
            )
            energy_response.components.extend([cpu_component, gpu_component])

            energy_response.batch_size = entry[3]
            energy_response_list.append(energy_response)

    return energy_response_list



import enum
import logging

import torch

from deepview_profile.tracking.backward_interceptor import BackwardInterceptor
from deepview_profile.tracking.breakdown import HierarchicalBreakdownBuilder
from deepview_profile.tracking.memory.activations import ActivationsTracker
from deepview_profile.tracking.memory.report import MemoryReportBuilder, MiscSizeType
from deepview_profile.tracking.memory.weights import WeightsTracker
from deepview_profile.tracking.time.operation import OperationRunTimeTracker
from deepview_profile.tracking.time.report import OperationRunTimeReportBuilder
from deepview_profile.user_code_utils import user_code_environment

logger = logging.getLogger(__name__)


class Tracker:
    def __init__(
        self,
        model_provider,
        iteration_provider,
        input_provider,
        project_root,
        user_code_path,
    ):
        self._model_provider = model_provider
        self._iteration_provider = iteration_provider
        self._input_provider = input_provider
        self._project_root = project_root
        self._user_code_path = user_code_path

        # Overall tracker state - used for correctness. Essentially, *if* the
        # caller wants to perform memory tracking, they must do so first before
        # run time tracking.
        self._tracker_state = _TrackerState.CREATED

        # Shared state
        self._model = None

        # Memory tracking state
        self._weight_tracker = None
        self._activations_tracker = None
        self._peak_usage_bytes = None

        # Run time tracking state
        self._operation_tracker = None

    def track_memory(self):
        if self._tracker_state != _TrackerState.CREATED:
            raise RuntimeError(
                'Memory tracking must be the first operation performed '
                'on a new Tracker.'
            )

        self._tracker_state = _TrackerState.MEMORY_TRACKED

        # Sanity check: For accurate profiling numbers, the initial memory
        # usage should be zero bytes.
        initial_memory_bytes = torch.cuda.memory_allocated()
        if initial_memory_bytes != 0:
            logger.debug(
                'Non-zero initial memory usage during memory tracking: '
                '%d bytes',
                initial_memory_bytes,
            )

        # 1. Track and record memory usage associated with model creation
        self._weight_tracker = WeightsTracker(self._project_root)
        with user_code_environment(self._user_code_path, self._project_root):
            with self._weight_tracker.track():
                self._model = self._model_provider()
            # Run one iteration to initialize the gradients
            iteration = self._iteration_provider(self._model)
            iteration(*self._input_provider())

        # 2. Track and record memory usage associated with stored activations
        self._activations_tracker = ActivationsTracker(self._project_root)
        self._activations_tracker.track_memory_usage(
            iteration, self._input_provider, self._user_code_path)

        # 3. Record peak memory usage
        torch.cuda.reset_max_memory_allocated()
        with user_code_environment(self._user_code_path, self._project_root):
            iteration(*(self._input_provider()))
        self._peak_usage_bytes = torch.cuda.max_memory_allocated()

    def track_run_time(self):
        if self._tracker_state == _TrackerState.CREATED:
            with user_code_environment(
                    self._user_code_path, self._project_root):
                self._model = self._model_provider()
        elif self._tracker_state != _TrackerState.MEMORY_TRACKED:
            raise RuntimeError('Run time tracking has already been performed.')

        self._tracker_state = _TrackerState.RUN_TIME_TRACKED

        # 2. Perform operation run time profiling
        with user_code_environment(self._user_code_path, self._project_root):
            inputs = self._input_provider()
            iteration = self._iteration_provider(self._model)

        print("Tracking both input and iteration")
        self._operation_tracker = OperationRunTimeTracker(self._project_root)
        backward_interceptor = BackwardInterceptor()
        with self._operation_tracker.track():
            with backward_interceptor.intercept():
                with user_code_environment(
                        self._user_code_path, self._project_root):
                    inputs = self._input_provider()
                    iteration(*inputs)

    def get_memory_report(self, report_file=None):
        if (self._weight_tracker is None or
                self._activations_tracker is None or
                self._peak_usage_bytes is None):
            raise RuntimeError('Memory tracking has not been performed yet.')

        return (MemoryReportBuilder(report_file)
                .process_tracker(self._weight_tracker)
                .process_tracker(self._activations_tracker)
                .add_misc_entry(
                    MiscSizeType.PeakUsageBytes, self._peak_usage_bytes)
                .build())

    def get_run_time_report(self, report_file=None):
        if self._operation_tracker is None:
            raise RuntimeError('Run time tracking has not been performed yet.')

        return (OperationRunTimeReportBuilder(report_file)
                .process_tracker(self._operation_tracker)
                .build())

    def get_hierarchical_breakdown(self):
        if (self._weight_tracker is None or
                self._activations_tracker is None or
                self._peak_usage_bytes is None or
                self._operation_tracker is None):
            raise RuntimeError(
                'Memory tracking and run time tracking have not both been '
                'performed yet.'
            )

        return (HierarchicalBreakdownBuilder()
                .for_model(self._model)
                .set_peak_usage_bytes(self._peak_usage_bytes)
                .process_tracker(self._operation_tracker)
                .process_tracker(self._activations_tracker)
                .process_tracker(self._weight_tracker)
                .build())


class _TrackerState(enum.Enum):
    CREATED = 0
    MEMORY_TRACKED = 1
    RUN_TIME_TRACKED = 2


import collections
import math
import os

HierarchicalBreakdown = collections.namedtuple(
    'HierarchicalBreakdown', ['operations', 'weights', 'peak_usage_bytes'])


class HierarchicalBreakdownBuilder:
    def __init__(self):
        self._module_names_by_id = {}
        self._operation_root = OperationNode('root', -1)
        self._weight_root = WeightNode('root', -1)
        self._peak_usage_bytes = None

    def for_model(self, model):
        for name, module in model.named_modules():
            self._module_names_by_id[id(module)] = name
        return self

    def process_tracker(self, tracker):
        tracker.populate_breakdown(self)
        return self

    def add_run_time_entry(
            self, operation_name, forward_ms, backward_ms, stack_context):
        if len(stack_context.frames) == 0:
            raise ValueError(
                'Adding run time entry with no context to the breakdown.')

        for entry in self._traverse_and_insert(
                self._operation_root, operation_name, stack_context):
            entry.add_run_time(forward_ms, backward_ms)

        return self

    def add_activation_entry(self, operation_name, size_bytes, stack_context):
        if len(stack_context.frames) == 0:
            raise ValueError(
                'Adding activation entry with no context to the breakdown.')

        for entry in self._traverse_and_insert(
                self._operation_root, operation_name, stack_context):
            entry.add_activation_size(size_bytes)

        return self

    def add_weight_entry(
            self, weight_name, size_bytes, grad_size_bytes, stack_context):
        if len(stack_context.frames) == 0:
            raise ValueError(
                'Adding weight entry with no context to the breakdown.')

        for entry in self._traverse_and_insert(
                self._weight_root, weight_name, stack_context):
            entry.add_weight_size(size_bytes, grad_size_bytes)

        return self

    def set_peak_usage_bytes(self, peak_usage_bytes):
        self._peak_usage_bytes = peak_usage_bytes
        return self

    def build(self):
        if self._peak_usage_bytes is None:
            raise RuntimeError(
                'Missing peak usage when constructing the breakdown.')

        self._prune_tree(self._operation_root)
        self._prune_tree(self._weight_root)
        self._operation_root.build_context_info_map()
        return HierarchicalBreakdown(
            operations=self._operation_root,
            weights=self._weight_root,
            peak_usage_bytes=self._peak_usage_bytes,
        )

    def _traverse_and_insert(self, root, leaf_name, stack_context):
        """
        A generator that, given a list of relevant stack frames, traverses (and
        inserts entries, if needed) the hierarchical breakdown tree, yielding
        each node along its path.
        """
        parent = root
        node_constructor = type(root)
        stack_frames = stack_context.frames

        for idx, frame in enumerate(reversed(stack_frames)):
            is_last_frame = idx == len(stack_frames) - 1
            context = (frame.file_path, frame.line_number)

            if context not in parent.children:
                name = (
                    leaf_name if is_last_frame
                    else self._module_names_by_id[
                        stack_frames[-idx-2].module_id]
                )
                if name == '':
                    name = 'Model'
                new_entry = node_constructor(name, frame.module_id)
                new_entry.add_context(context)
                parent.children[context] = new_entry

            yield parent
            parent = parent.children[context]

        parent.add_name(leaf_name)
        yield parent

    def _prune_tree(self, root):
        # Current node, key to node from parent, parent node
        stack = [(root, None, None)]

        # Depth first traversal. Prune interior nodes with only one child:
        #   e.g. ... -> parent -> node -> child -> ... becomes
        #        ... -> parent -> child -> ...
        while len(stack) != 0:
            node, key, parent = stack.pop()

            if len(node.children) == 1 and parent is not None:
                # Remove "node" from the tree and have the parent
                # point directly to node's only child
                child = next(iter(node.children.values()))
                child.add_context(key)
                parent.children[key] = child
                node.children.clear()
                stack.append((child, key, parent))
            else:
                for key_to_child, child in node.children.items():
                    stack.append((child, key_to_child, node))


class BreakdownNode:
    def __init__(self, name, module_id):
        self._names = [name]
        self._module_id = module_id

        self._children = {}
        self._contexts = []

    def add_context(self, context):
        self._contexts.append(context)

    def add_name(self, name):
        if name in self._names:
            return
        self._names.append(name)

    @property
    def name(self):
        return ', '.join(self._names)

    @property
    def children(self):
        return self._children

    def serialize_data_to_protobuf(self, entry):
        # Template method, must be implemented by subclasses
        raise NotImplementedError

    def serialize_to_protobuf(self, array):
        # Serialize using a preorder traversal
        stack = [self]
        while len(stack) != 0:
            node = stack.pop()

            entry = array.add()
            entry.name = node.name
            entry.num_children = len(node.children)
            for context in reversed(node._contexts):
                file_ref = entry.contexts.add()
                _serialize_file_ref(file_ref, context)
            node.serialize_data_to_protobuf(entry)

            for child in node.children.values():
                stack.append(child)


class OperationNode(BreakdownNode):
    def __init__(self, name, module_id):
        super().__init__(name, module_id)
        self._forward_ms = 0.
        self._backward_ms = None
        self._size_bytes = 0
        self._context_info_map = None

    def add_run_time(self, forward_ms, backward_ms):
        self._forward_ms += forward_ms

        if backward_ms is None:
            return
        if self._backward_ms is None:
            self._backward_ms = 0.
        self._backward_ms += backward_ms

    def add_activation_size(self, size_bytes):
        self._size_bytes += size_bytes

    @property
    def forward_ms(self):
        return self._forward_ms

    @property
    def backward_ms(self):
        return self._backward_ms

    @property
    def size_bytes(self):
        return self._size_bytes

    def serialize_data_to_protobuf(self, entry):
        entry.operation.forward_ms = self.forward_ms
        entry.operation.backward_ms = (
            self.backward_ms if self.backward_ms is not None
            else math.nan
        )
        entry.operation.size_bytes = self.size_bytes
        for context, context_info in self._context_info_map.items():
            map_entry = entry.operation.context_info_map.add()
            _serialize_file_ref(map_entry.context, context)
            context_info.serialize_to_protobuf(map_entry)

    def build_context_info_map(self):
        """
        Builds aggregate memory/run time usage information for each tracked
        line of code.
        """
        stack = [(self, 0)]
        while len(stack) > 0:
            node, visit_count = stack.pop()

            if visit_count > 0:
                for child in node.children.values():
                    context_info = ContextInfo(
                        size_bytes=child.size_bytes,
                        run_time_ms=(
                            child.forward_ms if child.backward_ms is None
                            else child.forward_ms + child.backward_ms
                        ),
                    )
                    for context in child._contexts:
                        if context in node._context_info_map:
                            node._context_info_map[context] += context_info
                        else:
                            node._context_info_map[context] = context_info

                    ContextInfo.merge_map(
                        node._context_info_map,
                        child._context_info_map,
                    )
            else:
                node._context_info_map = {}
                stack.append((node, 1))
                for child in node.children.values():
                    stack.append((child, 0))


class WeightNode(BreakdownNode):
    def __init__(self, name, module_id):
        super().__init__(name, module_id)
        self._size_bytes = 0
        self._grad_size_bytes = 0

    def add_weight_size(self, size_bytes, grad_size_bytes):
        self._size_bytes += size_bytes
        self._grad_size_bytes += grad_size_bytes

    @property
    def size_bytes(self):
        return self._size_bytes

    @property
    def grad_size_bytes(self):
        return self._grad_size_bytes

    def serialize_data_to_protobuf(self, entry):
        entry.weight.size_bytes = self.size_bytes
        entry.weight.grad_size_bytes = self.grad_size_bytes


class ContextInfo:
    def __init__(self, size_bytes, run_time_ms, invocations=1):
        self._size_bytes = size_bytes
        self._run_time_ms = run_time_ms
        self._invocations = invocations

    def __add__(self, other):
        return ContextInfo(
            size_bytes=self.size_bytes + other.size_bytes,
            run_time_ms=self.run_time_ms + other.run_time_ms,
            invocations=self.invocations + other.invocations,
        )

    def __iadd__(self, other):
        return self.__add__(other)

    def __repr__(self):
        return (
            'ContextInfo(invocations={:d}, size_bytes={:d}, '
            'run_time_ms={:.4f})'.format(
                self.invocations, self.size_bytes, self.run_time_ms)
        )

    @property
    def size_bytes(self):
        return self._size_bytes

    @property
    def run_time_ms(self):
        return self._run_time_ms

    @property
    def invocations(self):
        return self._invocations

    @staticmethod
    def merge_map(destination, to_merge):
        for key, value in to_merge.items():
            if key in destination:
                destination[key] += value
            else:
                destination[key] = value

    def serialize_to_protobuf(self, entry):
        entry.run_time_ms = self.run_time_ms
        entry.size_bytes = self.size_bytes
        entry.invocations = self.invocations


def _serialize_file_ref(file_ref, context):
    file_path, line_number = context
    file_ref.line_number = line_number
    file_ref.file_path.components.extend(file_path.split(os.sep))



import sys


def print_analysis_error(error, file=sys.stderr):
    print(
        "DeepView encountered an error when profiling your model:",
        file=file,
    )
    print("->", str(error), file=file)

    if error.file_context is not None:
        if error.file_context.line_number is not None:
            message = (
                "This error occurred on line {} when processing {}.".format(
                    error.file_context.line_number,
                    error.file_context.file_path,
                )
            )
        else:
            message = "This error occurred when processing {}.".format(
                error.file_context.file_path,
            )
        print("->", message, file=file)



import datetime
import os
import sqlite3

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DB_PATH = os.path.join(BASE_DIR, "deepview.sqlite3")

class DatabaseInterface:
    def __init__(self, database_name=DB_PATH) -> None:
        self.connection = sqlite3.connect(
            database_name, detect_types=sqlite3.PARSE_DECLTYPES|sqlite3.PARSE_COLNAMES
        )
        self.create_energy_table()

    def create_energy_table(self) -> None:
        self.connection.cursor().execute("CREATE TABLE IF NOT EXISTS ENERGY ( \
            entry_point TEXT, \
            cpu_component REAL, \
            gpu_component REAL, \
            batch_size INT, \
            ts TIMESTAMP \
        );")


class EnergyTableInterface:
    def __init__(self, database_connection: sqlite3.Connection):
        self.database_connection: sqlite3.Connection = database_connection

    @staticmethod
    def is_valid_entry(entry: list) -> bool:
        '''
        Validates an entry in the Energy table by testing if the length is 3,
        and the types match the columns. Note that timestamp is not part of the entry.
        Returns True if it is valid, else False
        '''
        return len(entry) == 4 and type(entry[0]) == str and type(entry[1]) == float \
            and type(entry[2]) == float and type(entry[3]) == int

    @staticmethod
    def is_valid_entry_with_timestamp(entry: list) -> bool:
        '''
        Validates an entry in the Energy table by testing if the length is 4,
        and the types match the columns. Returns True if it is valid, else False
        '''
        return len(entry) == 5 and type(entry[0]) == str and type(entry[1]) == float \
            and type(entry[2]) == float and type(entry[3]) == int \
            and type(entry[4]) == datetime.datetime

    def add_entry(self, entry: list) -> bool:
        '''
        Validates an entry and then adds that entry into the Energy table. Note that
        current timestamp is added by this function. Returns False if the entry is
        not a valid format, or if the insertion failed. Else returns True
        '''
        if self.is_valid_entry(entry):
            try:
                entry.append(datetime.datetime.now())
                cursor = self.database_connection.cursor()
                cursor.execute("INSERT INTO ENERGY VALUES(?, ?, ?, ?, ?)", entry)
                self.database_connection.commit()
                return True
            except sqlite3.IntegrityError as e:
                print(e)
                return False
        else:
            return False

    def get_latest_n_entries_of_entry_point(self, n: int, entry_point: str) -> list:
        '''
        Gets the n latest entries of a given entry point
        '''
        params = [entry_point, n]
        cursor = self.database_connection.cursor()
        results = cursor.execute(
            "SELECT * FROM ENERGY WHERE entry_point=? ORDER BY ts DESC LIMIT ?;",
            params
        ).fetchall()
        return results



import torch
import logging
import gc
import os
import base64
from google.protobuf.json_format import MessageToDict

logger = logging.getLogger(__name__)

def release_memory():
    logger.debug("Emptying cache")
    gc.collect()
    torch.cuda.empty_cache()

def next_message_to_dict(object):
    message = next(object)
    return MessageToDict(message)

def files_encoded_unique(operation_tree):
    encoded_files = []

    for analysis in operation_tree:
        context_info_map = analysis["operation"].get("contextInfoMap", None)
        if context_info_map is not None and len(context_info_map) > 0:
            filename = list(
                context_info_map[0]["context"]["filePath"]["components"]
            ).pop()

            already_in_list = next(
                (item for item in encoded_files if item["name"] == filename), None
            )
            if not already_in_list:
                file_path = os.path.join(
                    "", *list(context_info_map[0]["context"]["filePath"]["components"])
                )

                encoded_file = encode_file("", file_path)
                encoded_files.append(encoded_file)

    return encoded_files

def encode_file(root, file):
    file_dict = None
    if os.path.splitext(file)[1] == ".py" and file != "entry_point.py":
        file_dict = {"name": file, "content": ""}

        filename = os.path.join(root, file)

        with open(filename, "r") as f:
            file_content = f.read()
            file_dict["content"] = base64.b64encode(
                file_content.encode("utf-8")
            ).decode("utf-8")

    return file_dict

def model_location_patterns():
    return [
        "./transformers/models[/\w+/]+\w+.py",
        "./transformers/integrations[/\w+/]+\w+.py",
        "./diffusers/models[/\w+/]+\w+.py",
    ]


import contextlib
import torch

from deepview_profile.exceptions import _SuspendExecution
from deepview_profile.tracking.hook_manager import HookManager


class BackwardInterceptor:
    def __init__(self):
        self._backward_hooks = HookManager()
        self.backward_root = None

    @contextlib.contextmanager
    def intercept(self):
        self._backward_hooks.attach_hook(
            torch.Tensor,
            'backward',
            self._hook_creator,
        )
        try:
            yield
        except _SuspendExecution:
            pass
        finally:
            self._backward_hooks.remove_hooks()

    def _hook_creator(self, fn):
        def hook(*args):
            self.backward_root = args[0]
            raise _SuspendExecution
        return hook



import argparse
import logging
import os
from deepview_profile.analysis.session import AnalysisSession
from deepview_profile.nvml import NVML
from deepview_profile.utils import release_memory
import weakref

def analyze_project(project_root, entry_point, nvml, enable_ddp_analysis):
    session = AnalysisSession.new_from(project_root, entry_point)
    
    release_memory()
    print("analyze_project: running measure_breakdown()")
    yield session.measure_breakdown(nvml)
    
    release_memory()
    print("analyze_project: running measure_throughput()")
    yield session.measure_throughput()
    
    release_memory()
    print("analyze_project: running deepview_predict()")
    yield session.habitat_predict()
    
    release_memory()
    print("analyze_project: running measure_utilization()")
    yield session.measure_utilization()

    release_memory()
    print("analyze_project: running energy_compute()")
    yield session.energy_compute()

    if enable_ddp_analysis:
        release_memory()
        print("analyze_project: running ddp_computation()")
        yield session.ddp_computation()
    
    # release object session (less gpu memory consumption)
    release_memory()
    weakref.finalize(session, print, "session object destroyed")
    del session
    yield None

def main():
    # This is used for development and debugging purposes
    parser = argparse.ArgumentParser()
    parser.add_argument("entry_point", type=str)
    args = parser.parse_args()

    project_root = os.getcwd()
    with NVML() as nvml:
        analyzer = analyze_project(project_root, args.entry_point, nvml)
        breakdown = next(analyzer)
        throughput = next(analyzer)

    print('Peak usage:   ', breakdown.peak_usage_bytes, 'bytes')
    print('Max. capacity:', breakdown.memory_capacity_bytes, 'bytes')
    print('No. of weight breakdown nodes:   ', len(breakdown.operation_tree))
    print('No. of operation breakdown nodes:', len(breakdown.weight_tree))
    print('Throughput:', throughput.samples_per_second, 'samples/s')


if __name__ == "__main__":
    kwargs = {
        "format": "%(asctime)s %(levelname)-8s %(message)s",
        "datefmt": "%Y-%m-%d %H:%M",
        "level": logging.DEBUG,
    }
    logging.basicConfig(**kwargs)
    main()



import logging
import signal
import threading

from deepview_profile.initialization import (
    check_skyline_preconditions,
    initialize_skyline,
)

logger = logging.getLogger(__name__)


def register_command(subparsers):
    parser = subparsers.add_parser(
        "interactive",
        help="Start a new DeepView interactive profiling session.",
    )
    parser.add_argument(
        "--host",
        default="",
        help="The host address to bind to.",
    )
    parser.add_argument(
        "--port",
        default=60120,
        type=int,
        help="The port to listen on.",
    )
    parser.add_argument(
        "--hints-file",
        help="Path to the performance hints configuration YAML file.",
    )
    parser.add_argument(
        "--measure-for",
        help="Number of iterations to measure when determining throughput.",
    )
    parser.add_argument(
        "--warm-up",
        help="Number of warm up iterations when determining throughput.",
    )
    parser.add_argument(
        "--log-file",
        help="The location of the log file.",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Log debug messages.")
    parser.set_defaults(func=main)

def actual_main(args):
    from deepview_profile.server import SkylineServer

    should_shutdown = threading.Event()

    def signal_handler(signal, frame):
        should_shutdown.set()

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)


    with SkylineServer(args.host, args.port) as server:
        _, port = server.listening_on
        logger.info(
            "DeepView interactive profiling session started! "
            "Listening on port %d.",
            port,
        )

        # Run the server until asked to terminate
        should_shutdown.wait()


def main(args):
    check_skyline_preconditions(args)
    initialize_skyline(args)
    actual_main(args)



import logging 
import os
import sys
import json
import platform

from deepview_profile.analysis.runner import analyze_project
from deepview_profile.nvml import NVML
from deepview_profile.utils import release_memory, next_message_to_dict, files_encoded_unique

from deepview_profile.initialization import (
    check_skyline_preconditions,
    initialize_skyline,
)
from deepview_profile.error_printing import print_analysis_error

logger = logging.getLogger(__name__)

def register_command(subparsers):
    parser = subparsers.add_parser(
        "analysis",
        help="Generate usage report for various analysis.",
    )
    parser.add_argument(
        "entry_point",
        help="The entry point file in this project that contains the DeepView "
             "provider functions."
    )
    parser.add_argument(
        "--all",
        action="store_true",
        help="The complete analysis of all methods"
    )
    parser.add_argument(
        "-breakdown", "--measure-breakdown",
        action="store_true",
        help="Adds breakdown data to results"
    )
    parser.add_argument(
        "-throughput", "--measure-throughput",
        action="store_true",
        help="Adds throughput data to results"
    )
    parser.add_argument(
        "-predict", "--habitat-predict",
        action="store_true",
        help="Adds habitat data prediction to results"
    )
    parser.add_argument(
        "-utilization", "--measure-utilization",
        action="store_true",
        help="Adds utilization data to results"
    )
    parser.add_argument(
        "-energy", "--energy-compute",
        action="store_true",
        help="Adds energy use to results"
    )
    parser.add_argument(
        "--include-ddp",
        action="store_true",
        help="Adds ddp analysis to results"
    )
    parser.add_argument(
        "-o", "--output",
        help="The location where the complete report should be stored",
        required=True
    )
    parser.add_argument(
        "--log-file",
        help="The location of the log file",
    )
    parser.add_argument(
        "--exclude-source",
        action="store_true",
        help="Allows not adding encodedFiles section"
    )
    parser.add_argument("--debug", action="store_true", help="Log debug messages.")
    parser.set_defaults(func=main)

def measure_breakdown(session, nvml):
    print("analysis: running measure_breakdown()")
    yield session.measure_breakdown(nvml)
    release_memory()

def measure_throughput(session):
    print("analysis: running measure_throughput()")
    yield session.measure_throughput()
    release_memory()

def habitat_predict(session):
    print("analysis: running deepview_predict()")
    yield session.habitat_predict()
    release_memory()

def measure_utilization(session):
    print("analysis: running measure_utilization()")
    yield session.measure_utilization()
    release_memory()

def energy_compute(session):
    print("analysis: running energy_compute()")
    yield session.energy_compute()
    release_memory()

def ddp_analysis(session):
    print("analysis: running ddp_computation()")
    yield session.ddp_computation()
    release_memory()

def hardware_information(nvml):
    
    hardware_info = { 
        'hostname': platform.node(),
        'os': " ".join(list(platform.uname())),
        'gpus': nvml.get_device_names()
    }
    return hardware_info

def actual_main(args): 
    from deepview_profile.analysis.session import AnalysisSession
    from deepview_profile.exceptions import AnalysisError

    if os.path.exists(args.output):
        print(
            "ERROR: The specified output file already exists.",
            file=sys.stderr,
        )
        sys.exit(1)

    try: 
        project_root = os.getcwd()
        data = { 
            "analysisState": {
                "message_type": "analysis",
                "project_root": project_root,
                "project_entry_point": args.entry_point,
                "hardware_info": {},
                "throughput": {},
                "breakdown": {},
                "habitat": {},
                "additionalProviders": "",
                "energy": {},
                "utilization": {},
                "ddp": {}
            },
            "epochs": 50,
            "iterPerEpoch": 1000,
            "encodedFiles": []
        }

        session = AnalysisSession.new_from(project_root, args.entry_point)
        release_memory()

        is_return_all = args.all

        with NVML() as nvml:
            data['analysisState']['hardware_info'] = hardware_information(nvml)
            if args.measure_breakdown or is_return_all:
                data['analysisState']['breakdown'] = next_message_to_dict(measure_breakdown(session, nvml))

                operation_tree = data['analysisState']['breakdown']['operationTree']
                if not args.exclude_source and operation_tree is not None:
                    data['encodedFiles'] = files_encoded_unique(operation_tree)

        if args.measure_throughput or is_return_all:
            data['analysisState']['throughput'] = next_message_to_dict(measure_throughput(session))

        if args.habitat_predict or is_return_all:
            data['analysisState']['habitat'] = next_message_to_dict(habitat_predict(session))

        if args.measure_utilization or is_return_all:
            data['analysisState']['utilization'] = next_message_to_dict(measure_utilization(session))

        if args.energy_compute or is_return_all:
            data['analysisState']['energy'] = next_message_to_dict(energy_compute(session))

        if args.include_ddp:
            data['analysisState']['ddp'] = next_message_to_dict(ddp_analysis(session))

        with open(args.output, "w") as json_file:
            json.dump(data, json_file, indent=4)

    except AnalysisError as ex: 
        print_analysis_error(ex)
        sys.exit(1)

def main(args): 
    check_skyline_preconditions(args)
    initialize_skyline(args)
    actual_main(args)


import subprocess
import functools
import orjson
import io
import traceback
import numpy as np
import logging
import time
import os
from collections import deque
from perfetto.trace_processor import TraceProcessor
from torch_tb_profiler.profiler.tensor_core import TC_Allowlist
from torch.profiler import profile, schedule, ProfilerActivity

logger = logging.getLogger(__name__)
FILENAME = "raw_trace_file.json"


class Node:
    def __init__(
        self,
        name,
        start,
        end,
        duration,
        track,
        depth,
        slice_id,
        cpu_forward,
        gpu_forward,
        cpu_backward,
        gpu_backward,
    ):
        self.name = name
        self.start = start
        self.end = end
        self.duration = duration
        self.track = track
        self.depth = depth
        self.slice_id = slice_id
        self.cpu_forward = cpu_forward
        self.gpu_forward = gpu_forward
        self.gpu_forward_span = 0
        self.cpu_backward = cpu_backward
        self.gpu_backward = gpu_backward
        self.cpu_backward_span = 0
        self.cpu_backward_slices = list()
        self.gpu_backward_span = 0
        self.gpu_backward_slices = list()
        self.children = list()

    def __str__(self):
        return f"{self.name} \t {self.slice_id} "

    def postorder(self):
        ret = []
        for c in self.children:
            ret.extend(c.postorder())
        ret.append(self)

        return ret


class UtilizationProfiler:
    def __init__(self):
        self._root_node = None
        self._tensor_core_perc = None

    def _calculate_gpu_times(self, kernel_list):
        # return (span, net time)
        if not kernel_list:
            return 0, 0
        start, end, netTime = (
            kernel_list[0][0],
            kernel_list[0][1],
            kernel_list[0][1] - kernel_list[0][0],
        )

        for s, e in kernel_list[1:]:
            if start <= s < end and end < e:
                netTime += e - end
            elif end <= s:
                netTime += e - s
            end = max(end, e)

        return (end - start), netTime

    def _calculate_gpu_forward_time(self, tp, node):
        span = 0
        time = 0
        cudaLaunchList = tp.query_dict(
            f"""select * from slices where name like '%CudaLaunchKernel%' and track_id={node.track}
                                    and depth>{node.depth} and ts between {node.start} and {node.end} ORDER BY ts ASC"""
        )

        # add individual kernels
        kernel_list = []
        for cudaLaunch in cudaLaunchList:
            slice_id_origin = cudaLaunch["slice_id"]
            slice_id_destination = tp.query_dict(
                f"""select * from flow where slice_out={slice_id_origin}"""
            )
            if slice_id_destination:
                cuda_slice = tp.query_dict(
                    f"select * from slice where slice_id={slice_id_destination[0]['slice_in']}"
                )
                kernel_list.append(
                    (cuda_slice[0]["ts"], cuda_slice[0]["ts"] + cuda_slice[0]["dur"])
                )

        if kernel_list:
            kernel_list.sort(key=lambda x: x[0])
            span, time = self._calculate_gpu_times(kernel_list)

        node.gpu_forward_span = span
        node.gpu_forward = time

    def _backward_slices(self, tp):
        res = []
        backwardTopOps = tp.query_dict(
            "select * from slices where name like '%Autograd::engine%' and depth=0 ORDER BY ts ASC"
        )
        for bto in backwardTopOps:
            kernel_slices = []
            endTime = bto["ts"] + bto["dur"]
            cudaLaunchList = tp.query_dict(
                f"""select * from slices where name like '%CudaLaunchKernel%' and track_id={bto['track_id']}
                                    and depth>{bto['depth']} and ts between {bto['ts']} and {endTime}"""
            )

            for cudaLaunch in cudaLaunchList:
                slice_id_origin = cudaLaunch["slice_id"]
                slice_id_destination = tp.query_dict(
                    f"""select * from flow where slice_out={slice_id_origin}"""
                )
                if slice_id_destination:
                    cuda_slice = tp.query_dict(
                        f"select * from slice where slice_id={slice_id_destination[0]['slice_in']}"
                    )
                    kernel_slices.append(cuda_slice[0])

            bto["name"] = bto["name"].split()[-1]
            bto["kernel_list"] = kernel_slices
            res.append(bto)
        return res

    def _accumulate_backward_slices_to_node(self, node):
        for ch in node.children:
            node.cpu_backward_slices.extend(
                self._accumulate_backward_slices_to_node(ch)
            )

        return node.cpu_backward_slices

    def _calculate_backward_cpu_span(self, node):
        backward_slices = node.cpu_backward_slices
        kernel_slices = []
        if not backward_slices:
            return
        minTs = float("inf")
        maxTs = float("-inf")
        netTime = 0
        backward_slices.sort(key=lambda x: x["ts"])
        for bw_slice in backward_slices:
            minTs = min(minTs, bw_slice["ts"])
            maxTs = max(maxTs, bw_slice["ts"] + bw_slice["dur"])
            netTime += bw_slice["dur"]
            kernel_slices.extend(
                [
                    (kernel["ts"], kernel["ts"] + kernel["dur"])
                    for kernel in bw_slice["kernel_list"]
                ]
            )

        node.cpu_backward_span = maxTs - minTs
        node.cpu_backward = netTime
        if kernel_slices:
            kernel_slices.sort(key=lambda x: x[0])
            node.gpu_backward_span, node.gpu_backward = self._calculate_gpu_times(
                kernel_slices
            )

    def _populate_backward_data(self, node):
        self._calculate_backward_cpu_span(node)
        for ch in node.children:
            self._populate_backward_data(ch)

    @functools.lru_cache(maxsize=None)
    def _can_match(self, f, b):
        if "aten" in f and "Backward0" in b:
            raw_f = f[len("aten::") :].lower().replace("_", "")
            raw_b = b[: -len("Backward0")].lower()
            return (
                raw_f == raw_b
                or raw_f == "transpose"
                and raw_b == "t"
                or raw_f == "t"
                and raw_b == "transpose"
            )

        return False

    # solves the longest common subsequence (LCS) problem, matching
    # the post-ordered model forward pass with the sequence of backward
    # operators executed.
    #
    # this function returns a list of tuples whose elements are in the form:
    #  (forward, backward)

    def _lcs(self, forward, backward):
        N, M = len(forward), len(backward)
        dp = np.zeros((N, M))
        for i in range(N):
            for j in range(M):
                dp[i, j] = int(self._can_match(forward[i].name, backward[j]["name"]))
                if i > 0:
                    dp[i, j] = max(dp[i, j], dp[i - 1, j])
                if j > 0:
                    dp[i, j] = max(dp[i, j], dp[i, j - 1])
                if (
                    i > 0
                    and j > 0
                    and self._can_match(forward[i].name, backward[j]["name"])
                ):
                    dp[i, j] = max(dp[i, j], 1 + dp[i - 1, j - 1])

        matchings = []
        i, j = N - 1, M - 1
        while i > 0 and j > 0:
            if self._can_match(forward[i].name, backward[j]["name"]):
                matchings.append((forward[i], backward[j]))
                i -= 1
                j -= 1
            elif i > 0 and dp[i - 1, j] == dp[i, j]:
                i -= 1
            else:
                j -= 1

        if logging.root.level == logging.DEBUG:
            logger.debug(f"N = {N}, M = {M}, best matching: {len(matchings)}\n")
        return matchings

    def _convert_ids_int_string(self, slices):
        for slice in slices:
            if "id" in slice:
                slice["id"] = str(slice["id"])

    def _convert_negative_tids_to_positive(self, slices):
        for slice in slices:
            if "tid" in slice and isinstance(slice["tid"], int):
                slice["tid"] = abs(slice["tid"])

    def _remove_args(self, slices):
        [slice.pop("args", None) for slice in slices]

    def _filter_traces(self, raw_slices):
        names_to_filter = [
            "profiler.py",
            "built-in",
            "torch/",
            "cudaDeviceSynchronize",
            "typing.py",
            "<module>",
            "os.py",
            "_collections",
            "enum.py",
            "numpy/",
            "DataParallel",
            "lib/",
            ".py",
        ]
        idx_to_filter = []
        for idx, item in enumerate(raw_slices["traceEvents"]):
            # SKIP HIGH LEVEL TRACE IDENTIFIER FOR BACKWARD PASS
            if item["name"].startswith("torch/autograd") and item["name"].endswith(
                "backward"
            ):
                continue
            for keyword in names_to_filter:
                if keyword in item["name"]:
                    idx_to_filter.append(idx)
            # SPECIFIC CASE
            if item["name"].startswith("entry_point") and item["name"].endswith(
                "iteration"
            ):
                idx_to_filter.append(idx)
            if item["name"].startswith("entry_point") and item["name"].endswith(
                "forward"
            ):
                idx_to_filter.append(idx)

        filtered_slices = [
            event
            for idx, event in enumerate(raw_slices["traceEvents"])
            if idx not in idx_to_filter
        ]
        raw_slices["traceEvents"] = filtered_slices

        if logging.root.level == logging.DEBUG:
            with open("filtered_slices.json", "wb") as f:
                f.write(orjson.dumps(raw_slices))

    def _get_perfetto_object(self, filepath):
        with open(filepath, "rb") as f:
            raw_slices = orjson.loads(f.read())

        self._filter_traces(raw_slices)

        if isinstance(raw_slices, dict) and "traceEvents" in raw_slices:
            # Perfetto doesn't want this format produced by PyTorch
            raw_slices = raw_slices.pop("traceEvents", None)
        # Convert IDs from int to string. Without this perfetto fails to JSON load trace with IDs stored as integers.
        self._convert_ids_int_string(raw_slices)
        # Convert negative 'tid' values to positive. Without this perfetto combines together the slices with different tids into one track
        self._convert_negative_tids_to_positive(raw_slices)
        self._remove_args(raw_slices)  # For speedup

        slices_bytes = orjson.dumps(raw_slices)
        slices_bytes = io.BytesIO(slices_bytes)  # 4x speedup using orjson

        try:
            tp = TraceProcessor(slices_bytes)
        except ConnectionResetError as e:
            # This happens sometimes so retry once
            tp = TraceProcessor(slices_bytes)

        interesting_fields = "SELECT ts, dur, track_id, category, name, depth, cat, slice_id, id, arg_set_id FROM slice"

        def query_dict(query):
            query = query.lower().replace(
                "select * from slice", interesting_fields
            )  # gives a 15% speedup
            try:
                query_iterator = tp.query(query)
            except Exception as e:
                print("[ERROR] Unable to run query: %s" % query)
                print(traceback.format_exc())
                print("\n\n")
                raise e
            return [item.__dict__ for item in query_iterator]

        tp.query_dict = query_dict

        return tp

    def _convert_node_to_dict(self, node):
        newEntry = {
            "slice_id": node.slice_id,
            "name": node.name,
            "start": node.start,
            "end": node.end,
            "cpu_forward": node.duration,
            "cpu_forward_span": node.cpu_forward,
            "gpu_forward": node.gpu_forward,
            "gpu_forward_span": node.gpu_forward_span,
            "cpu_backward": node.cpu_backward,
            "cpu_backward_span": node.cpu_backward_span,
            "gpu_backward": node.gpu_backward,
            "gpu_backward_span": node.gpu_backward_span,
            "children": list(),
        }
        for ch in node.children:
            newEntry["children"].append(self._convert_node_to_dict(ch))
        return newEntry

    def output_to_json(self, path_save_file):
        node_json = self._convert_node_to_dict(self._root_node)
        output = {"node": node_json, "tensor_core": self._tensor_core_perc}
        with open(os.path.join(path_save_file, "profiling_results.json"), "wb") as f:
            f.write(orjson.dumps(output))

    def _calculate_tensor_core_utilization(self, filepath):
        kernelDict = {"tensorTime": 0, "noTensorTime": 0, "totalTime": 0}
        with open(filepath, "r") as f:
            data = orjson.loads(f.read())
        for event in data["traceEvents"]:
            if event.get("cat") and event["cat"] == "kernel":
                if event["name"] in TC_Allowlist:
                    kernelDict["tensorTime"] += event["dur"]
                else:
                    kernelDict["noTensorTime"] += event["dur"]

        totalTime = kernelDict["tensorTime"] + kernelDict["noTensorTime"]
        if logging.root.level == logging.DEBUG:
            logger.debug(
                f'Tensor time: {kernelDict["tensorTime"]} perc {round(kernelDict["tensorTime"]/totalTime*100,2)}\n'
            )
            logger.debug(
                f'No Tensor time: {kernelDict["noTensorTime"]} perc {round(kernelDict["noTensorTime"]/totalTime*100,2)}\n'
            )
            logger.debug(f"totalTime: {totalTime}\n")
        return round(kernelDict["tensorTime"] / totalTime * 100, 2)

    def _deepview_analysis(self, filepath):
        startTime = time.time()
        tp = self._get_perfetto_object(filepath)

        profilerStepStart = tp.query_dict(
            "select * from slices where name like '%ProfilerStep%'"
        )
        main_track = profilerStepStart[0]["track_id"]
        start = profilerStepStart[0]["ts"]
        end = start + profilerStepStart[0]["dur"]
        profilerStartDepth = profilerStepStart[0]["depth"]

        rootQuery = tp.query_dict(
            f"""
                                select * from slices where name like '%nn.Module:%' 
                                and depth = 
                                (SELECT MIN(depth) from slices where name like '%nn.Module%' and depth>{profilerStartDepth} and ts between {start} and {end} and track_id = {main_track})                
                                """
        )[0]

        rootNode = Node(
            rootQuery["name"],
            rootQuery["ts"],
            rootQuery["ts"] + rootQuery["dur"],
            rootQuery["dur"],
            rootQuery["track_id"],
            rootQuery["depth"],
            rootQuery["slice_id"],
            rootQuery["dur"],
            0,
            0,
            0,
        )
        self._calculate_gpu_forward_time(tp, rootNode)
        logger.debug(f"{rootNode}\n")

        stack = deque([rootNode])

        while stack:
            node = stack.popleft()
            queryNewDepth = tp.query_dict(
                f"""select MIN(depth) from slices where (name like '%nn.Module:%' or name like '%aten::%') and 
                                        depth>{node.depth} and track_id={node.track}
                                        and ts between {node.start} and {node.end}"""
            )
            minDepth = queryNewDepth[0]["min(depth)"] if queryNewDepth else -1
            if minDepth:
                queryResults = tp.query_dict(
                    f"""select * from slices where (name like '%nn.Module:%' or name like '%aten::%') 
                                        and depth={minDepth} and track_id={node.track} 
                                            and ts between {node.start} and {node.end}"""
                )
                for qr in queryResults:
                    newNode = Node(
                        qr["name"],
                        qr["ts"],
                        qr["ts"] + qr["dur"],
                        qr["dur"],
                        qr["track_id"],
                        qr["depth"],
                        qr["slice_id"],
                        qr["dur"],
                        0,
                        0,
                        0,
                    )
                    self._calculate_gpu_forward_time(tp, newNode)
                    node.children.append(newNode)
                    stack.append(newNode)

        ForwardPostOrderTraversal = list(
            filter(lambda x: "aten::" in x.name, rootNode.postorder())
        )
        ForwardPostOrderTraversal.sort(key=lambda x: x.start, reverse=True)
        backwardSlices = self._backward_slices(tp)
        matchings = self._lcs(ForwardPostOrderTraversal, backwardSlices)

        countAccumulateGrad = 0
        for slice in backwardSlices:
            if "AccumulateGrad" in slice["name"]:
                countAccumulateGrad += 1

        numValidBackwardSlices = len(backwardSlices) - countAccumulateGrad
        if logging.root.level == logging.DEBUG:
            logger.debug(f"number of valid slices: {numValidBackwardSlices}\n")
            logger.debug(
                f"Number of matched slices: {len(matchings)} percentage: {round(len(matchings)/numValidBackwardSlices*100,2)}\n"
            )

        for match in matchings:
            node = match[0]
            backward_slice = match[1]
            node.cpu_backward_slices.append(backward_slice)

        ## DEBUGGING PROCESS BACKWARD SLICES MATCHING #############
        if logging.root.level == logging.DEBUG:
            matchings.sort(key=lambda x: x[1]["ts"])
            with open("matching_results.txt", "w") as file:
                file.write(f"Total number of backward slices: {len(backwardSlices)}\n")
                file.write(f"number of valid slices: {numValidBackwardSlices}\n")
                file.write(
                    f"Number of matched slices: {len(matchings)} percentage: {round(len(matchings)/numValidBackwardSlices*100,2)}\n"
                )

                for m in matchings:
                    file.write(
                        f"forward: {m[0].name} {m[0].slice_id} backward: {m[1]['name']} {m[1]['slice_id']}\n"
                    )

        self._accumulate_backward_slices_to_node(rootNode)
        self._populate_backward_data(rootNode)

        if logging.root.level == logging.DEBUG:
            profilingResult = self._convert_node_to_dict(rootNode)
            with open("profiling_results.json", "wb") as f:
                f.write(orjson.dumps(profilingResult))

        endTime = time.time()
        logger.debug(f"Total elapsed time: {endTime - startTime}")
        self._root_node = rootNode
        self._tensor_core_perc = self._calculate_tensor_core_utilization(filepath)

        # DELETE PYTORCH PROFILER TRACE (except for debugging purposes)
        if not (logging.root.level == logging.DEBUG):
            subprocess.run(["rm", "-f", os.path.join(os.getcwd(), filepath)])


def _trace_handler(p):
    p.export_chrome_trace(FILENAME)


def _serialize_node(respNode, internalNode):
    respNode.slice_id = internalNode["slice_id"]
    respNode.name = internalNode["name"]
    respNode.start = int(internalNode["start"])
    respNode.end = int(internalNode["end"])
    respNode.cpu_forward = int(internalNode["cpu_forward"])
    respNode.cpu_forward_span = int(internalNode["cpu_forward_span"])
    respNode.gpu_forward = int(internalNode["gpu_forward"])
    respNode.gpu_forward_span = int(internalNode["gpu_forward_span"])
    respNode.cpu_backward = int(internalNode["cpu_backward"])
    respNode.cpu_backward_span = int(internalNode["cpu_backward_span"])
    respNode.gpu_backward = int(internalNode["gpu_backward"])
    respNode.gpu_backward_span = int(internalNode["gpu_backward_span"])

    for ch in internalNode["children"]:
        addRespNode = respNode.children.add()
        _serialize_node(addRespNode, ch)


def serialize_response(respNode, rootNode):
    _serialize_node(respNode, rootNode)


def utilization_analysis(model_provider, input_provider, iteration_provider):
    model = model_provider()
    inputs = input_provider()
    iteration = iteration_provider(model)
    skip_first = 2
    wait = 1
    warmup = 1
    active = 1
    totalIterations = skip_first + wait + warmup + active
    deepviewSchedule = schedule(
        skip_first=skip_first, wait=wait, warmup=warmup, active=active, repeat=1
    )

    start = time.time()
    elapsed = 0

    while elapsed < 30:
        for _ in range(100):
            iteration(*inputs)
        elapsed = time.time() - start

    with profile(
        activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
        schedule=deepviewSchedule,
        on_trace_ready=_trace_handler,
        with_stack=True,
    ) as p:
        for _ in range(totalIterations):
            iteration(*inputs)
            p.step()

    utilization = UtilizationProfiler()
    path_to_file = os.path.join(os.getcwd(), FILENAME)
    utilization._deepview_analysis(path_to_file)
    jsonFormat = {
        "root_node": utilization._convert_node_to_dict(utilization._root_node),
        "tensor_core_perc": utilization._tensor_core_perc,
    }
    return jsonFormat





class LRUCache:
    def __init__(self, max_size=128):
        self._max_size = max_size
        self._cache_by_key = {}
        self._cache_by_use = _LRUCacheList()

    def query(self, key):
        if key not in self._cache_by_key:
            return None
        node = self._cache_by_key[key]
        self._cache_by_use.move_to_front(node)
        return node.value

    def add(self, key, value):
        if self._cache_by_use.size >= self._max_size:
            removed = self._cache_by_use.remove_back()
            del self._cache_by_key[removed.key]
        node = self._cache_by_use.add_to_front(key, value)
        self._cache_by_key[key] = node


class _LRUCacheList:
    def __init__(self):
        # Front of the list: most recently used
        self.front = None
        self.back = None
        self.size = 0

    def add_to_front(self, key, value):
        node = _LRUCacheNode(key, value)
        self._add_to_front(node)
        self.size += 1
        return node

    def _add_to_front(self, node):
        if self.size == 0:
            self.front = node
            self.back = node
        else:
            node.next = self.front
            self.front.prev = node
            self.front = node

    def move_to_front(self, node):
        if self.front == node:
            # Nothing needs to be done if the node is already at the front of
            # the list
            return

        if node.next is None:
            # Back of the list
            node.prev.next = None
            self.back = node.prev
            node.prev = None
        else:
            # Middle of the list
            node.prev.next = node.next
            node.next.prev = node.prev
            node.next = None
            node.prev = None

        self._add_to_front(node)

    def remove_back(self):
        if self.size == 0:
            return None

        node = self.back

        if self.size == 1:
            self.front = None
            self.back = None
        else:
            node.prev.next = None
            self.back = node.prev

        self.size -= 1
        return node


class _LRUCacheNode:
    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.prev = None
        self.next = None



import logging
import socket

from deepview_profile.io.connection import Connection, ConnectionState
from deepview_profile.exceptions import NoConnectionError

logger = logging.getLogger(__name__)


class ConnectionManager:
    def __init__(self, message_handler, closed_handler):
        self._connections = {}
        self._message_handler = message_handler
        self._closed_handler = closed_handler

    def register_connection(self, opened_socket, address):
        self._connections[address] = (
            Connection(
                opened_socket,
                address,
                self._message_handler,
                self._closed_handler,
            ),
            ConnectionState(),
        )
        self._connections[address][0].start()

    def remove_connection(self, address):
        connection, state = self.get_connection_tuple(address)
        connection.stop()
        state.connected = False
        del self._connections[address]
        logger.debug("Removed connection to (%s:%d).", *address)

    def get_connection(self, address):
        return self.get_connection_tuple(address)[0]

    def get_connection_state(self, address):
        return self.get_connection_tuple(address)[1]

    def get_connection_tuple(self, address):
        if address not in self._connections:
            host, port = address
            raise NoConnectionError(
                "Connection to ({}:{}) does not exist.".format(host, port))
        return self._connections[address]

    def broadcast(self, string_message):
        for _, (connection, _) in self._connections.items():
            connection.write_string_message(string_message)

    def connect_to(self, host, port):
        new_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        new_socket.connect((host, port))
        self.register_connection(new_socket, (host, port))

    def stop(self):
        for _, (connection, state) in self._connections.items():
            connection.stop()
            state.connected = False
        self._connections.clear()





class Position:
    def __init__(self, line, column):
        self.line = line
        self.column = column

    def offset(self, length):
        return Position(self.line, self.column + length)


class SourceMap:
    def __init__(self, source_code):
        self._source_by_line = source_code.splitlines()

    def find_position(self, snippet, line_offset=0):
        for offset, line in enumerate(self._source_by_line[line_offset:]):
            index = line.find(snippet)
            if index == -1:
                continue
            # NOTE: We don't add 1 here to make the line number 0-based
            return Position(line_offset + offset, index)

        return None

    def find_position_on_line(self, snippet, offset_position):
        if offset_position.line >= len(self._source_by_line):
            return None
        index = self._source_by_line[offset_position.line].find(
            snippet, offset_position.column)

        if index == -1:
            return None
        else:
            return Position(offset_position.line, index)



import yaml

import deepview_profile.data


class _Config:
    def __init__(self):
        self.Hints = None

        self.warm_up = 100
        self.measure_for = 10

    def initialize_hints_config(self, hints_file):
        if hints_file is None:
            file_to_open = deepview_profile.data.get_absolute_path('hints.yml')
        else:
            file_to_open = hints_file

        with open(file_to_open, 'r') as f:
            self.Hints = yaml.load(f, Loader=yaml.Loader)

    def parse_args(self, args):
        if 'hints_file' not in args:
            args.hints_file = None
        self.initialize_hints_config(args.hints_file)

        if 'warm_up' in args and args.warm_up is not None:
            self.warm_up = args.warm_up
        if 'measure_for' in args and args.measure_for is not None:
            self.measure_for = args.measure_for


Config = _Config()



import logging
import select
import socket
from threading import Thread

from deepview_profile.io.sentinel import Sentinel

logger = logging.getLogger(__name__)


class ConnectionAcceptor:
    """
    Manages the "server socket" for the agent, allowing it to accept
    connection requests from other agents.

    Each time a connection is received, the handler_function is called
    with the new socket and address.
    """
    def __init__(self, host, port, handler_function):
        self._host = host
        self._port = port
        self._server_socket = socket.socket(
            socket.AF_INET,
            socket.SOCK_STREAM,
        )
        self._server_socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_REUSEADDR,
            1,
        )
        self._handler_function = handler_function
        self._acceptor = Thread(target=self._accept_connections)
        self._sentinel = Sentinel()

    def start(self):
        self._server_socket.bind((self._host, self._port))
        self._port = self._server_socket.getsockname()[1]
        self._server_socket.listen()
        self._sentinel.start()
        self._acceptor.start()
        logger.debug(
            "DeepView is listening for connections on (%s:%d).",
            self._host,
            self._port,
        )

    def stop(self):
        self._sentinel.signal_exit()
        self._acceptor.join()
        self._server_socket.close()
        self._sentinel.stop()
        logging.debug(
            "DeepView has stopped listening for connections on (%s:%d).",
            self._host,
            self._port,
        )

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    def _accept_connections(self):
        try:
            while True:
                read_ready, _, _ = select.select(
                    [self._server_socket, self._sentinel.read_pipe], [], [])

                if self._sentinel.should_exit(read_ready):
                    self._sentinel.consume_exit_signal()
                    break

                socket, address = self._server_socket.accept()
                host, port = address
                logger.debug("Accepted a connection to (%s:%d).", host, port)
                self._handler_function(socket, address)
        except Exception:
            logging.exception(
                "DeepView has unexpectedly stopped accepting connections.")



import collections

from deepview_profile.tracking.base import ReportBase, ReportBuilderBase
import deepview_profile.tracking.time.report_queries as queries

RunTimeEntry = collections.namedtuple(
    'RunTimeEntry',
    ['operation_name',
     'forward_ms',
     'backward_ms',
     'file_path',
     'line_number'],
)


class OperationRunTimeReport(ReportBase):
    def __init__(self, connection):
        super().__init__(connection)

    def get_run_time_entries(self, path_prefix=None):
        cursor = self._connection.cursor()
        return map(
            lambda row: RunTimeEntry(*row),
            cursor.execute(queries.get_run_time_entries_with_context),
        )


class OperationRunTimeReportBuilder(ReportBuilderBase):
    # This is the operation run time tracking report file format version that
    # will be created by this builder. When changes are made to the file
    # format, this integer should be increased monotonically.
    #
    # We need to version these tracking reports to protect us from future
    # changes to the file format.
    Version = 1

    def __init__(self, file=None):
        super().__init__(file)

    def add_run_time_entry(
            self, operation_name, forward_ms, backward_ms, stack_context):
        cursor = self._connection.cursor()
        cursor.execute(queries.add_run_time_entry, (
            operation_name,
            forward_ms,
            backward_ms,
        ))
        entry_id = cursor.lastrowid

        def stack_frame_generator():
            for idx, frame in enumerate(stack_context.frames):
                yield (idx, frame.file_path, frame.line_number, entry_id)

        cursor.executemany(queries.add_stack_frame, stack_frame_generator())

    def build(self):
        self._connection.commit()
        return OperationRunTimeReport(self._connection)

    def _create_report_tables(self):
        cursor = self._connection.cursor()
        cursor.execute(queries.set_report_format_version.format(
            version=OperationRunTimeReportBuilder.Version))
        for creation_query in queries.create_report_tables.values():
            cursor.execute(creation_query)
        self._connection.commit()



import torch
import logging

from deepview_profile.profiler.autograd import AutogradEngine
from deepview_profile.profiler.backward import BackwardHelper, backward_available

logger = logging.getLogger(__name__)


class OperationProfiler:
    def __init__(self, warm_up=3, measure_for=3):
        self._warm_up = warm_up
        self._measure_for = measure_for
        self._start_event = torch.cuda.Event(enable_timing=True)
        self._end_event = torch.cuda.Event(enable_timing=True)

    def measure_operation_ms(self, func, args, kwargs):
        for_inplace = _is_potentially_inplace(func.__name__)

        forward_args, forward_kwargs = self._get_args_for_profiling(
            args, kwargs, for_inplace)
        def forward_runnable():
            func(*forward_args, **forward_kwargs)
        forward_ms = self._measure_ms(forward_runnable)

        # We need separate copies of the arguments for the forward and backward
        # measurements because func might be inplace. Running an inplace
        # function repeatedly will affect the autograd graph, which causes
        # problems when we try to measure the backward pass.
        backward_args, backward_kwargs = self._get_args_for_profiling(
            args, kwargs, for_inplace)
        retval = func(*backward_args, **backward_kwargs)
        if not backward_available(retval):
            return forward_ms, None

        return forward_ms, self._measure_backward_ms(func.__name__, retval)

    def _get_args_for_profiling(self, args, kwargs, for_inplace=False):
        cloned_args = tuple(map(
            lambda arg: self._clone_tensors(arg, for_inplace), args))
        cloned_kwargs = {
            key: self._clone_tensors(value, for_inplace)
            for key, value in kwargs.items()
        }
        return cloned_args, cloned_kwargs

    def _clone_tensors(self, argument, for_inplace):
        if isinstance(argument, torch.Tensor):
            detached = argument.detach()
            detached.requires_grad_(argument.requires_grad)
            # We need to clone the tensor for inplace operations because they
            # cannot be executed on a leaf tensor. This adds some overhead to
            # our backward measurements (an extra CloneBackward function), but
            # it _should_ be negligible. I chose not to exclude CloneBackward
            # from the backward measurements to avoid introducing incorrectness
            # if the user actually uses clone() in their own code.
            return detached if not for_inplace else detached.clone()

        if isinstance(argument, tuple):
            return tuple(map(
                lambda arg: self._clone_tensors(arg, for_inplace), argument))

        if isinstance(argument, list):
            return list(map(
                lambda arg: self._clone_tensors(arg, for_inplace), argument))

        return argument

    def _measure_backward_ms(self, func_name, operation_outputs):
        # As of PyTorch 1.5.1, sometimes our AutogradEngine will not work
        # because the behavior of grad_functions has changed. When this
        # happens, we fall back to using PyTorch's existing backward engine.
        #
        # The reason we do not always use the existing backward engine is that
        # the existing engine has a start up overhead that is non-negligible
        # when there are many short operations involved.
        try:
            return self._measure_backward_engine_strategy(operation_outputs)
        except (RuntimeError, TypeError):
            logger.debug("%s: Falling back to PyTorch's engine", func_name)
            return self._measure_backward_torch_strategy(operation_outputs)

    def _measure_backward_engine_strategy(self, operation_outputs):
        engine = AutogradEngine.new_from(operation_outputs)
        return self._measure_ms(engine.run_backward)

    def _measure_backward_torch_strategy(self, operation_outputs):
        helper = BackwardHelper.new_from(operation_outputs)

        backward_ms = self._measure_ms(helper.run_backward)
        accum_grad_ms = self._measure_ms(helper.run_accumulate_grad)
        diff = backward_ms - accum_grad_ms

        return diff if diff >= 1e-6 else backward_ms

    def _measure_ms(self, runnable):
        for _ in range(self._warm_up):
            runnable()

        self._start_event.record()
        for _ in range(self._measure_for):
            runnable()
        self._end_event.record()
        torch.cuda.synchronize()

        return (
            self._start_event.elapsed_time(self._end_event) / self._measure_for
        )


# Populated manually from:
# https://pytorch.org/docs/stable/nn.functional.html
POTENTIALLY_INPLACE_FUNCTIONS = {
    'threshold',
    'relu',
    'hardtanh',
    'relu6',
    'elu',
    'selu',
    'celu',
    'leaky_relu',
    'rrelu',
    'dropout',
    'alpha_dropout',
    'dropout2d',
    'dropout3d',

    # In place math operations (+=, *=, -=, /=, //=)
    '__iadd__',
    '__imul__',
    '__isub__',
    '__itruediv__',
    '__ifloordiv__',
}


def _is_potentially_inplace(fn_name):
    return (
        fn_name in POTENTIALLY_INPLACE_FUNCTIONS or
        # In PyTorch, functions with a '_' suffix are in place, by convention
        (len(fn_name) > 1 and fn_name[-1] == '_' and fn_name[-2] != '_')
    )



import collections
import contextlib
import os
import traceback
import logging

logger = logging.getLogger(__name__)

FileContext = collections.namedtuple(
    'FileContext',
    ['file_path', 'line_number'],
)


class AnalysisError(RuntimeError):
    def __init__(self, message, exception_type=None):
        if exception_type is None:
            super(AnalysisError, self).__init__(message)
        else:
            super(AnalysisError, self).__init__(
                '{}: {}'.format(exception_type.__name__, message))

        self.file_context = None

    def with_file_context(self, file_path, line_number=None):
        self.file_context = FileContext(
            file_path=file_path,
            line_number=line_number,
        )
        return self


class NoConnectionError(Exception):
    def __init__(self, message):
        super().__init__(message)


class _SuspendExecution(Exception):
    # This exception is used internally by the BackwardInterceptor to return
    # early from the user's code.
    pass


@contextlib.contextmanager
def exceptions_as_analysis_errors(project_root):
    try:
        yield
    except _SuspendExecution:
        # A _SuspendExecution exception does not indicate an error - we use it
        # to return early from the user's code.
        pass
    except AnalysisError:
        # The user's code may raise an AnalysisError (e.g., from the wrapped
        # providers). If this happens, we should pass the exception through.
        raise
    except Exception as ex:
        logger.debug(
            "An error occured during analysis (could be a problem with the "
            "user's code):",
            exc_info=ex,
        )
        if isinstance(ex, SyntaxError):
            error = AnalysisError(
                'DeepView encountered a syntax error while profiling your '
                'model.'
            )
        else:
            error = AnalysisError(str(ex), type(ex))

        # Extract the relevant file context, if it is available, starting by
        # inspecting the exception itself.
        if hasattr(ex, 'filename') and ex.filename.startswith(project_root):
            _add_context_to_error(
                error, project_root, ex.filename, getattr(ex, 'lineno', None))
        else:
            stack = traceback.extract_tb(ex.__traceback__)
            for frame in reversed(stack):
                if frame.filename.startswith(project_root):
                    _add_context_to_error(
                        error, project_root, frame.filename, frame.lineno)
                    break

        # Special case: Add a more detailed error message when there's an
        # input number mismatch.
        if (error.file_context is None and
                str(error).startswith("TypeError: forward() takes")):
            error = AnalysisError(
                "{}. This error could be due to a mismatch between the number "
                "of inputs that your model expects and the number of inputs "
                "that your input provider returns.".format(str(error))
            )

        raise error


def _add_context_to_error(error, project_root, file_path, line_number):
    error.with_file_context(
        file_path=os.path.relpath(file_path, start=project_root),
        line_number=line_number,
    )



import os

import toml

file_path = os.path.abspath(__file__)
dir_path = os.path.dirname(file_path)
package = toml.load(os.path.join(dir_path, "..", "pyproject.toml"))

__name__ = package["tool"]["poetry"]["name"]
__version__ = package["tool"]["poetry"]["version"]
__description__ = package["tool"]["poetry"]["description"]

from .__main__ import main


import os
import logging
import pynvml
import platform

from deepview_profile.exceptions import NoConnectionError

import deepview_profile.protocol_gen.innpv_pb2 as pm

logger = logging.getLogger(__name__)


class MessageSender:
    def __init__(self, connection_manager):
        self._connection_manager = connection_manager

    def send_initialize_response(self, context):
        message = pm.InitializeResponse()
        connection = self._connection_manager.get_connection(context.address)
        message.server_project_root = connection.project_root
        message.entry_point.components.extend(connection.entry_point.split(os.sep))

        # Populate hardware info
        message.hardware.hostname = platform.node()
        message.hardware.os = " ".join(list(platform.uname()))
        pynvml.nvmlInit()
        for i in range(pynvml.nvmlDeviceGetCount()):
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            device_name = pynvml.nvmlDeviceGetName(handle).decode("utf-8")
            message.hardware.gpus.append(device_name)
        pynvml.nvmlShutdown()

        self._send_message(message, 'initialize', context)

    def send_protocol_error(self, error_code, context):
        message = pm.ProtocolError()
        message.error_code = error_code
        self._send_message(message, 'error', context)

    def send_breakdown_response(self, breakdown, context):
        # Ideally, MessageSender users should not need to know about the INNPV
        # protocol messages. However, to avoid extraneous copies, sometimes
        # callers will pass in constructed messages for sending.
        self._send_message(breakdown, 'breakdown', context)

    def send_analysis_error(self, analysis_error, context):
        message = pm.AnalysisError()
        message.error_message = str(analysis_error)
        if analysis_error.file_context is not None:
            message.file_context.file_path.components.extend(
                analysis_error.file_context.file_path.split(os.sep)
            )
            message.file_context.line_number = (
                analysis_error.file_context.line_number
                if analysis_error.file_context.line_number is not None
                else 0
            )
        self._send_message(message, 'analysis_error', context)

    def send_throughput_response(self, throughput, context):
        self._send_message(throughput, 'throughput', context)

    def send_habitat_response(self, habitat_resp, context):
        self._send_message(habitat_resp, 'habitat', context)

    def send_energy_response(self, energy_resp, context):
        self._send_message(energy_resp, 'energy', context)

    def send_utilization_response(self, utilization_resp, context):
        self._send_message(utilization_resp, 'utilization', context)

    def send_ddp_response(self, ddp_resp, context):
        self._send_message(ddp_resp, 'ddp', context)

    def _send_message(self, message, payload_name, context):
        try:
            connection = self._connection_manager.get_connection(
                context.address)
            enclosing_message = pm.FromServer()
            getattr(enclosing_message, payload_name).CopyFrom(message)
            enclosing_message.sequence_number = context.sequence_number 
            connection.send_bytes(enclosing_message.SerializeToString())
        except NoConnectionError:
            logger.debug(
                'Not sending message to (%s:%d) because it is no longer '
                'connected.',
                *context.address,
            )



import torch


class BackwardHelper:
    def __init__(self, backward_runnable, ag_dict):
        self.run_backward = backward_runnable
        self._ag_dict = ag_dict

    @classmethod
    def new_from(cls, operation_outputs):
        retval, initial_grad_fn = get_grad_fn(operation_outputs)
        if initial_grad_fn is None:
            raise ValueError('No grad_fn available on the operation output.')

        grads = torch.ones_like(retval)
        def backward_runnable():
            torch.autograd.backward(retval, grads, retain_graph=True)

        size_dict = get_accumulate_grad_inputs(
            initial_grad_fn,
            backward_runnable,
        )

        ag_dict = {
            grad_fn: torch.randn(size, device=torch.device('cuda'))
            for grad_fn, size in size_dict.items()
        }

        return cls(backward_runnable, ag_dict)

    def run_accumulate_grad(self):
        for grad_fn, grad in self._ag_dict.items():
            grad_fn(grad)


def backward_available(operation_output):
    return get_grad_fn(operation_output)[1] is not None


def flatten_operation_output(operation_output):
    if isinstance(operation_output, torch.Tensor):
        return [operation_output]
    elif (not isinstance(operation_output, tuple) and
          not isinstance(operation_output, list)):
        return []

    flattened = []
    for value in operation_output:
        flattened.extend(flatten_operation_output(value))
    return flattened


def get_grad_fn(retval):
    if isinstance(retval, torch.Tensor) and retval.grad_fn is not None:
        return retval, retval.grad_fn
    elif isinstance(retval, tuple) or isinstance(retval, list):
        for inner_value in retval:
            inner_retval, grad_fn = get_grad_fn(inner_value)
            if grad_fn is not None:
                return inner_retval, grad_fn

    return None, None


def get_accumulate_grad_inputs(initial_grad_fn, backward_runnable):
    input_dict = {}
    hook_handles = []
    def get_hook(grad_fn):
        def hook(arg1, arg2):
            if not isinstance(arg2[0], torch.Tensor):
                return
            input_dict[grad_fn] = arg2[0].size()
        return hook

    # Traverse the graph to identify all AccumulateGrad functions
    stack = [initial_grad_fn]
    visited = {initial_grad_fn}

    while len(stack) > 0:
        grad_fn = stack.pop()

        if grad_fn.name() == 'torch::autograd::AccumulateGrad':
            hook_handles.append(grad_fn.register_hook(get_hook(grad_fn)))

        for next_grad_fn, _ in grad_fn.next_functions:
            if next_grad_fn is None or next_grad_fn in visited:
                continue
            stack.append(next_grad_fn)
            visited.add(next_grad_fn)

    # Run a backward pass to get accumulate grad sizes
    backward_runnable()
    torch.cuda.synchronize()

    # Clear hooks
    for handle in hook_handles:
        handle.remove()

    return input_dict



import logging
import sys

logger = logging.getLogger(__name__)


def check_skyline_preconditions(args):
    """
    This is the first function that should run before importing any other
    DeepView code.
    """
    _configure_logging(args)
    if not _validate_dependencies():
        sys.exit(1)
    if not _validate_gpu():
        sys.exit(1)


def initialize_skyline(args):
    """
    Performs common initialization tasks.
    """
    from deepview_profile.config import Config

    Config.parse_args(args)

def _configure_logging(args):
    kwargs = {
        "format": "%(asctime)s %(levelname)-8s %(message)s",
        "datefmt": "%Y-%m-%d %H:%M",
        "level": logging.DEBUG if args.debug else logging.INFO,
    }

    if args.log_file is not None:
        kwargs["filename"] = args.log_file

    logging.basicConfig(**kwargs)


def _validate_dependencies():
    # NOTE: If you make a change here, make sure to update the INSTALL_REQUIRES
    #       list in setup.py as well.
    try:
        import yaml # pyyaml on PyPI # noqa: F401
        import pynvml # nvidia-ml-py3 on PyPI # noqa: F401
        import google.protobuf # protobuf on PyPI # noqa: F401
        import numpy # noqa: F401
        import torch # noqa: F401
        return True
    except ImportError as ex:
        logger.error(
            "DeepView could not find the '%s' module, which is a required "
            "dependency. Please make sure all the required dependencies are "
            "installed before launching DeepView. If you use a package "
            "manager, these dependencies will be automatically installed for "
            "you.",
            ex.name,
        )
        return False


def _validate_gpu():
    import torch
    if not torch.cuda.is_available():
        logger.error(
            "DeepView did not detect a GPU on this machine. DeepView only "
            "profiles deep learning workloads on GPUs."
        )
        return False
    return True



import inspect

import torch

from deepview_profile.tracking.base import TrackerBase
from deepview_profile.tracking.hook_manager import HookManager
from deepview_profile.version_utils import Version

OLD_VF_PATH_VERSION = Version.parse_semantic_version('1.4.0')


class CallableTracker(TrackerBase):
    def __init__(self, hook_creator):
        super().__init__()
        self._hook_manager = HookManager()
        self._hook_creator = hook_creator
        self._torch_version = Version.parse_semantic_version(torch.__version__)

    def start_tracking(self):
        super().start_tracking()
        self._hook_manager.attach_hooks_on_module(
            torch,
            lambda fn: _is_callable_and_public(fn) and \
              fn.__name__ not in BLACKLISTED_TORCH_METHODS,
            self._hook_creator,
        )
        self._hook_manager.attach_hooks_on_module(
            torch.Tensor,
            lambda fn: _is_callable_and_public(fn) and \
              fn.__name__ != 'backward' and \
              fn.__name__ not in BLACKLISTED_TENSOR_METHODS,
            self._hook_creator,
        )
        self._hook_manager.attach_hooks_on_module(
            torch.Tensor,
            _is_callable_dunder,
            self._hook_creator,
        )
        self._hook_manager.attach_hooks_on_module(
            torch.nn.functional,
            _is_callable_and_public,
            self._hook_creator,
        )

        # The _VF module was moved to the torch module after version 1.4.0.
        # This is an unfortunate hack because we need to monkey patch certain
        # internal PyTorch functions to be able to identify all the operations
        # properly. The _VF module contains recurrent operations (e.g., lstm).
        vf_module = (
            torch._VF if self._torch_version is None or
                self._torch_version > OLD_VF_PATH_VERSION
            else torch.nn._VF
        )
        self._hook_manager.attach_hooks_on_module_using(
            vf_module,
            torch._C._VariableFunctions,
            _is_callable_and_public,
            self._hook_creator,
        )

    def stop_tracking(self):
        super().stop_tracking()
        self._hook_manager.remove_hooks()


def _is_callable_and_public(maybe_fn):
    # By convention, _ prefixed functions in Python should not be
    # called by users (i.e. they are "private" functions)
    return _is_callable(maybe_fn) and maybe_fn.__name__[0] != '_'

# Original source of these blacklists:
# https://github.com/NVIDIA/apex/blob/master/apex/pyprof/nvtx/nvmarker.py
BLACKLISTED_DUNDERS = {
    '__all__',
    '__array__',
    '__array_priority__',
    '__array_wrap__',
    '__bool__',
    '__builtins__',
    '__cached__',
    '__class__',
    '__deepcopy__',
    '__delattr__',
    '__delitem__',
    '__dict__',
    '__dir__',
    '__doc__',
    '__file__',
    '__format__',
    '__getattribute__',
    '__getitem__',
    '__hash__',
    '__index__',
    '__init__',
    '__init_subclass__',
    '__iter__',
    '__len__',
    '__loader__',
    '__module__',
    '__name__',
    '__new__',
    '__nonzero__',
    '__package__',
    '__path__',
    '__reduce__',
    '__reduce_ex__',
    '__repr__',
    '__reversed__',
    '__setattr__',
    '__setitem__',
    '__setstate__',
    '__sizeof__',
    '__spec__',
    '__str__',
    '__subclasshook__',
    '__version__',
    '__weakref__',
}

BLACKLISTED_TENSOR_METHODS = {
    'size', 'dim', 'item', 'tolist',
}

BLACKLISTED_TORCH_METHODS = {
    'is_storage',
}


def _is_callable_dunder(maybe_fn):
    """
    Returns True if maybe_fn is a callable dunder (callable named with double
    underscores) (e.g., __add__)
    """
    return (
        _is_callable(maybe_fn) and
        len(maybe_fn.__name__) > 4 and
        maybe_fn.__name__[:2] == '__' and
        maybe_fn.__name__[-2:] == '__' and
        maybe_fn.__name__ not in BLACKLISTED_DUNDERS
    )


def _is_callable(maybe_fn):
    return (
        inspect.isfunction(maybe_fn) or
        inspect.ismethod(maybe_fn) or
        inspect.isbuiltin(maybe_fn) or
        inspect.isroutine(maybe_fn)
    )



import collections
import re

Version = collections.namedtuple('Version', ['major', 'minor', 'patch'])

VERSION_REGEX = re.compile('^\d+\.\d+\.\d+$')


class Version:
    def __init__(self, major, minor, patch):
        self._major = major
        self._minor = minor
        self._patch = patch

    @property
    def major(self):
        return self._major

    @property
    def minor(self):
        return self._minor

    @property
    def patch(self):
        return self._patch

    def __repr__(self):
        return ''.join([
            'Version(major=',
            str(self.major),
            ', minor=',
            str(self.minor),
            ', patch=',
            str(self.patch),
            ')'
        ])

    def __eq__(self, other):
        return (
            self.major == other.major and
            self.minor == other.minor and
            self.patch == other.patch
        )

    def __gt__(self, other):
        self_nums = [self.major, self.minor, self.patch]
        other_nums = [other.major, other.minor, other.patch]

        for self_ver, other_ver in zip(self_nums, other_nums):
            if self_ver > other_ver:
                return True
            elif self_ver < other_ver:
                return False

        return False

    def __ne__(self, other):
        return not (self == other)

    def __ge__(self, other):
        return self == other or self > other

    def __lt__(self, other):
        return not (self >= other)

    def __le__(self, other):
        return not (self > other)

    @classmethod
    def parse_semantic_version(cls, version_str):
        if VERSION_REGEX.match(version_str) is None:
            return None
        version_nums = list(map(int, version_str.split('.')))
        return cls(
            major=version_nums[0],
            minor=version_nums[1],
            patch=version_nums[2],
        )



import logging
import math
import time
from concurrent.futures import ThreadPoolExecutor
from deepview_profile.analysis.runner import analyze_project
from deepview_profile.exceptions import AnalysisError
from deepview_profile.nvml import NVML
import deepview_profile.protocol_gen.innpv_pb2 as pm
import sys
logger = logging.getLogger(__name__)


class AnalysisRequestManager:
    def __init__(self, enqueue_response, message_sender, connection_manager):
        # We use a separate thread to perform the analysis request to not block
        # the server thread from handling messages from other connections.
        #
        # NOTE: We don't actually get parallelism because of the GIL.
        self._executor = ThreadPoolExecutor(max_workers=1)

        # Callable that enqueues a function onto the main executor
        self._enqueue_response = enqueue_response
        self._message_sender = message_sender
        self._connection_manager = connection_manager
        self._nvml = NVML()

    def start(self):
        self._nvml.start()

    def stop(self):
        self._nvml.stop()
        self._executor.shutdown()

    def submit_request(self, analysis_request, context):
        if analysis_request.mock_response:
            self._handle_mock_analysis_request(analysis_request, context)
            return

        self._executor.submit(
            self._handle_analysis_request,
            analysis_request,
            context,
        )

    def _handle_analysis_request(self, analysis_request, context):
        print("handle_analysis_request: begin")
        start_time = time.perf_counter()
        try:
            logger.debug(
                "Processing request %d from (%s:%d).",
                context.sequence_number,
                *(context.address),
            )
            connection = self._connection_manager.get_connection(context.address)
            analyzer = analyze_project(
                connection.project_root, 
                connection.entry_point, 
                self._nvml,
                analysis_request.ddp_analysis_request
            )

            # Abort early if the connection has been closed
            if self._early_disconnection_error(context):
                return

            breakdown = next(analyzer)
            self._enqueue_response(
                self._send_breakdown_response,
                breakdown,
                context,
            )

            if self._early_disconnection_error(context):
                return

            throughput = next(analyzer)
            self._enqueue_response(
                self._send_throughput_response,
                throughput,
                context,
            )

            # send habitat response
            if self._early_disconnection_error(context):
                return

            habitat_resp = next(analyzer)
            self._enqueue_response(
                self._send_habitat_response,
                habitat_resp,
                context,
            )

            # send utilization data
            if self._early_disconnection_error(context):
                return

            utilization_resp = next(analyzer)
            self._enqueue_response(
                self._send_utilization_response, utilization_resp, context
            )

            # send energy response
            if self._early_disconnection_error(context):
                return

            energy_resp = next(analyzer)
            self._enqueue_response(
                self._send_energy_response,
                energy_resp,
                context,
            )

            # send ddp analysis
            if self._early_disconnection_error(context):
                return
            
            if analysis_request.ddp_analysis_request:
                ddp_resp = next(analyzer)
                self._enqueue_response(
                    self._send_ddp_response,
                    ddp_resp,
                    context,
                )
            
            next(analyzer)

            # execution time
            elapsed_time = time.perf_counter() - start_time
            logger.debug(
                "Processed analysis request %d from (%s:%d) in %.4f seconds.",
                context.sequence_number,
                *(context.address),
                elapsed_time,
            )

        except AnalysisError as ex:
            self._enqueue_response(self._send_analysis_error, ex, context)

        except Exception:
            logger.exception("Exception occurred when handling analysis request.")
            self._enqueue_response(
                self._send_analysis_error,
                AnalysisError(
                    "An unexpected error occurred when analyzing your model. "
                    "Please file a bug report and then restart DeepView."
                ),
                context,
            )

    def _handle_mock_analysis_request(self, analysis_request, context):
        # This runs on the main executor
        breakdown = pm.BreakdownResponse()
        breakdown.peak_usage_bytes = 1337
        breakdown.memory_capacity_bytes = 13337
        breakdown.iteration_run_time_ms = 133.7
        self._message_sender.send_breakdown_response(breakdown, context)

        throughput = pm.ThroughputResponse()
        throughput.samples_per_second = 1337
        throughput.predicted_max_samples_per_second = math.nan
        self._message_sender.send_throughput_response(throughput, context)

    def _send_breakdown_response(self, breakdown, context):
        # Called from the main executor. Do not call directly!
        try:
            self._message_sender.send_breakdown_response(breakdown, context)
        except Exception:
            logger.exception("Exception occurred when sending a breakdown response.")

    def _send_analysis_error(self, exception, context):
        # Called from the main executor. Do not call directly!
        try:
            self._message_sender.send_analysis_error(exception, context)
        except Exception:
            logger.exception("Exception occurred when sending an analysis error.")

    def _send_throughput_response(self, throughput, context):
        # Called from the main executor. Do not call directly!
        try:
            self._message_sender.send_throughput_response(throughput, context)
        except Exception:
            logger.exception("Exception occurred when sending a throughput response.")

    def _send_habitat_response(self, habitat_resp, context):
        # Called from the main executor. Do not call directly!
        try:
            self._message_sender.send_habitat_response(habitat_resp, context)
        except Exception:
            logger.exception(
                "Exception occurred when sending a DeepView.Predict response."
            )

    def _send_energy_response(self, energy_resp, context):
        # Called from the main executor. Do not call directly!
        try:
            self._message_sender.send_energy_response(energy_resp, context)
        except Exception:
            logger.exception("Exception occurred when sending an energy response.")

    def _send_utilization_response(self, utilization_resp, context):
        # Called from the main executor. Do not call directly!
        try:
            self._message_sender.send_utilization_response(utilization_resp, context)
        except Exception:
            logger.exception("Exception occurred when sending utilization response.")

    def _send_ddp_response(self, ddp_response, context):
        # Called from the main executor. Do not call directly!
        try:
            self._message_sender.send_ddp_response(ddp_response, context)
        except Exception:
            logger.exception("Exception occurred when sending ddp response.")

    def _early_disconnection_error(self, context):
        if not context.state.connected:
            logger.error(
                "Aborting request %d from (%s:%d) early "
                "because the client has disconnected.",
                context.sequence_number,
                *(context.address),
            )
            return True

        return False



import logging
import os
import sys
import csv

from deepview_profile.initialization import (
    check_skyline_preconditions,
    initialize_skyline,
)
from deepview_profile.error_printing import print_analysis_error

logger = logging.getLogger(__name__)


def register_command(subparsers):
    parser = subparsers.add_parser(
        "prediction-models",
        help="Evaluate DeepView's prediction accuracy.",
    )
    parser.add_argument(
        "entry_point",
        help="The entry point file in this project that contains the DeepView "
             "provider functions.",
    )
    parser.add_argument(
        "-b", "--batch-sizes",
        help="The starting batch sizes to build models from.",
        type=int,
        nargs="+",
        required=True,
    )
    parser.add_argument(
        "-o", "--output",
        help="The location where the evaluation output should be stored.",
        required=True,
    )
    parser.add_argument(
        "--log-file",
        help="The location of the log file.",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Log debug messages.")
    parser.set_defaults(func=main)


def get_model(session, batch_size):
    # This is a HACK
    session._batch_size = batch_size
    thpt_msg = session.measure_throughput()
    return (
        (thpt_msg.peak_usage_bytes.slope, thpt_msg.peak_usage_bytes.bias),
        (thpt_msg.run_time_ms.slope, thpt_msg.run_time_ms.bias),
    )


def actual_main(args):
    from deepview_profile.analysis.session import AnalysisSession
    from deepview_profile.exceptions import AnalysisError

    if os.path.exists(args.output):
        print(
            "ERROR: The specified output file already exists.",
            file=sys.stderr,
        )
        sys.exit(1)

    try:
        with open(args.output, 'w') as f:
            writer = csv.writer(f)
            writer.writerow([
                'batch_size',
                'run_time_ms_slope',
                'run_time_ms_bias',
                'memory_usage_bytes_slope',
                'memory_usage_bytes_bias',
            ])
            project_root = os.getcwd()
            for batch_size in args.batch_sizes:
                session = AnalysisSession.new_from(
                    project_root, args.entry_point)
                memory_model, run_time_model = get_model(
                    session, batch_size)
                writer.writerow([
                    batch_size, *run_time_model, *memory_model,
                ])

    except AnalysisError as ex:
        print_analysis_error(ex)
        sys.exit(1)


def main(args):
    check_skyline_preconditions(args)
    initialize_skyline(args)
    actual_main(args)



import re


DUNDER_REGEX = re.compile('__(?P<name>.+)__')


def tensor_size_bytes(tensor):
    if tensor is None or not tensor.is_cuda:
        return 0
    return tensor.numel() * tensor.element_size()


def remove_dunder(fn_name):
    match = DUNDER_REGEX.match(fn_name)
    if match is None:
        return fn_name
    return match.group('name')



import collections
import logging
import os

from deepview_profile.exceptions import NoConnectionError

import deepview_profile.protocol_gen.innpv_pb2 as pm

logger = logging.getLogger(__name__)

RequestContext = collections.namedtuple(
    'RequestContext',
    ['address', 'state', 'sequence_number'],
)

def _validate_paths(project_root, entry_point):
    if not os.path.isabs(project_root):
        logger.error(
            "The project root that DeepView received is not an absolute path. "
            "This is an unexpected error. Please report a bug."
        )
        logger.error("Current project root: %s", project_root)
        return False

    if os.path.isabs(entry_point):
        logger.error(
            "The entry point must be specified as a relative path to the "
            "current directory. Please double check that the entry point you "
            "are providing is a relative path.",
        )
        logger.error("Current entry point path: %s", entry_point)
        return False

    full_path = os.path.join(project_root, entry_point)
    if not os.path.isfile(full_path):
        logger.error(
            "Either the specified entry point is not a file or its path was "
            "specified incorrectly. Please double check that it exists and "
            "that its path is correct.",
        )
        logger.error("Current absolute path to entry point: %s", full_path)
        return False

    return True


class MessageHandler:
    def __init__(
        self,
        connection_manager,
        message_sender,
        analysis_request_manager,
    ):
        self._connection_manager = connection_manager
        self._message_sender = message_sender
        self._analysis_request_manager = analysis_request_manager

    def _handle_initialize_request(self, message, context):
        if context.state.initialized:
            self._message_sender.send_protocol_error(
                pm.ProtocolError.ErrorCode.ALREADY_INITIALIZED_CONNECTION,
                context,
            )
            return
        if message.protocol_version != 5:
            # We only support version 5 of the protocol. We do not guarantee
            # backward compatibility for v0.x.x releases.
            # Version 1 - v0.1.x
            # Version 2 - v0.2.x
            # Version 3 - v0.3.x
            # Version 4 - v0.4.x
            self._message_sender.send_protocol_error(
                pm.ProtocolError.ErrorCode.UNSUPPORTED_PROTOCOL_VERSION,
                context,
            )
            self._connection_manager.remove_connection(context.address)
            logger.error(
                'DeepView is out of date. Please update to the latest versions '
                'of the DeepView command line interface and plugin.'
            )
            return


        if not _validate_paths(message.project_root,  message.entry_point):
            # Change this to the error related to
            self._message_sender.send_protocol_error(
                pm.ProtocolError.ErrorCode.UNSUPPORTED_PROTOCOL_VERSION,
                context,
            )
            self._connection_manager.remove_connection(context.address)
            logger.error(
                'Invalid project root or entry point.'
            )
            return
        logger.info("Connection addr:(%s:%d)", *context.address)
        logger.info("Project Root:   %s", message.project_root)
        logger.info("Entry Point:    %s", message.entry_point)
        self._connection_manager.get_connection(context.address)\
            .set_project_paths(message.project_root, message.entry_point)

        context.state.initialized = True
        self._message_sender.send_initialize_response(context)

    def _handle_analysis_request(self, message, context):
        if not context.state.initialized:
            self._message_sender.send_protocol_error(
                pm.ProtocolError.ErrorCode.UNINITIALIZED_CONNECTION,
                context,
            )
            return

        self._analysis_request_manager.submit_request(message, context)

    def handle_message(self, raw_data, address):
        try:
            message = pm.FromClient()
            message.ParseFromString(raw_data)
            logger.debug('Received message from (%s:%d).', *address)

            state = self._connection_manager.get_connection_state(address)
            if not state.is_request_current(message):
                logger.debug('Ignoring stale message from (%s:%d).', *address)
                return
            state.update_sequence(message)

            message_type = message.WhichOneof('payload')
            if message_type is None:
                logger.warn('Received empty message from (%s:%d).', *address)
                return

            context = RequestContext(
                state=state,
                address=address,
                sequence_number=message.sequence_number,
            )

            if message_type == 'initialize':
                self._handle_initialize_request(
                    getattr(message, message_type), context)
            elif message_type == 'analysis':
                self._handle_analysis_request(
                    getattr(message, message_type), context)
            else:
                # If the protobuf was compiled properly, this block should
                # never be reached.
                raise AssertionError(
                    'Invalid message type "{}".'.format(message_type))
        except NoConnectionError:
            logger.debug(
                'Dropping message from (%s:%d) because it is no longer '
                'connected.',
                *address,
            )
        except Exception:
            logger.exception(
                'Processing message from (%s:%d) resulted in an exception.',
                *address,
            )



import enum


class EntryType(enum.Enum):
    Weight = 1
    Activation = 2


create_report_tables = {
    'weight_entries': """
      CREATE TABLE IF NOT EXISTS weight_entries (
        id INTEGER PRIMARY KEY,
        name TEXT NOT NULL,
        size_bytes INTEGER NOT NULL,
        grad_size_bytes INTEGER NOT NULL
      )
    """,
    'activation_entries': """
      CREATE TABLE IF NOT EXISTS activation_entries (
        id INTEGER PRIMARY KEY,
        operation_name TEXT NOT NULL,
        size_bytes INTEGER NOT NULL
      )
    """,
    'correlation': """
      CREATE TABLE IF NOT EXISTS stack_correlation (
        correlation_id INTEGER PRIMARY KEY,
        entry_id INTEGER NOT NULL,
        entry_type INTEGER NOT NULL,
        UNIQUE (correlation_id, entry_id)
      )
    """,
    'correlation_index': """
      CREATE UNIQUE INDEX IF NOT EXISTS entry_type_and_id
        ON stack_correlation(entry_type, entry_id)
    """,
    'stack_frames': """
      CREATE TABLE IF NOT EXISTS stack_frames (
        correlation_id INTEGER NOT NULL,
        ordering INTEGER NOT NULL,
        file_path TEXT NOT NULL,
        line_number INTEGER NOT NULL,
        PRIMARY KEY (correlation_id, ordering)
      )
    """,
    'entry_types': """
      CREATE TABLE IF NOT EXISTS entry_types (
        entry_type INTEGER PRIMARY KEY,
        name TEXT NOT NULL
      )
    """,
    'misc_sizes': """
      CREATE TABLE IF NOT EXISTS misc_sizes (
        key TEXT PRIMARY KEY,
        size_bytes INT NOT NULL
      )
    """,
}

set_report_format_version = 'PRAGMA user_version = {version:d}'

add_entry_type = """
  INSERT INTO entry_types (entry_type, name) VALUES (?, ?)
"""

add_weight_entry = """
  INSERT INTO weight_entries (id, name, size_bytes, grad_size_bytes)
    VALUES (NULL, ?, ?, ?)
"""

add_activation_entry = """
  INSERT INTO activation_entries (id, operation_name, size_bytes)
    VALUES (NULL, ?, ?)
"""

add_correlation_entry = """
  INSERT INTO stack_correlation (correlation_id, entry_id, entry_type)
    VALUES (NULL, ?, ?)
"""

add_stack_frame = """
  INSERT INTO stack_frames (correlation_id, ordering, file_path, line_number)
    VALUES (?, ?, ?, ?)
"""

add_misc_entry = "INSERT INTO misc_sizes (key, size_bytes) VALUES (?, ?)"

get_misc_entry = "SELECT size_bytes FROM misc_sizes WHERE key = ?"

get_code_context_subquery = """
  WITH code_contexts AS (
    SELECT c.entry_id, s.file_path, s.line_number
      FROM stack_frames AS s JOIN stack_correlation AS c
        ON s.correlation_id == c.correlation_id
      WHERE
        c.entry_type = {:d}
      GROUP BY s.correlation_id HAVING s.ordering == MIN(s.ordering)
  )
"""

get_weight_entries_with_context = (
    get_code_context_subquery.format(EntryType.Weight.value) +
    """
      SELECT
          w.name, w.size_bytes, w.grad_size_bytes, c.file_path, c.line_number
        FROM weight_entries AS w
          LEFT JOIN code_contexts AS c
          ON w.id == c.entry_id
        WHERE w.size_bytes > 0
        ORDER BY c.file_path ASC, c.line_number ASC
    """
)

get_activation_entries_with_context = (
    get_code_context_subquery.format(EntryType.Activation.value) +
    """
      SELECT a.operation_name, a.size_bytes, c.file_path, c.line_number
        FROM activation_entries AS a
          LEFT JOIN code_contexts AS c
          ON a.id == c.entry_id
        WHERE a.size_bytes > 0
        ORDER BY c.file_path ASC, c.line_number ASC
    """
)



from torch.utils._pytree import tree_map
import torch.distributed as dist
import traceback
import orjson
from perfetto.trace_processor import TraceProcessor
import io
import torch
from typing import List, Tuple, Any

def get_bucket_sizes(model: torch.nn.Module, cap_size: int) -> List[int]:
    """
    Inputs: Pytorch model and a bucket size cap
    Outputs: list of bucket sizes in Megabytes.
    """
    params = [p for p in model.parameters() if p.requires_grad]
    bucket_cap_mb = cap_size
    bucket_bytes_cap = int(bucket_cap_mb * 1024 * 1024)
    bucket_size_limits = [
        dist._DEFAULT_FIRST_BUCKET_BYTES,
        bucket_bytes_cap,
    ]

    (
        bucket_indices,
        per_bucket_size_limits,
    ) = dist._compute_bucket_assignment_by_size(
        params, bucket_size_limits, [False] * len(params)
    )
    bucket_sizes = []
    bucket_indices_backward = bucket_indices[::-1]
    params_in_buckets = tree_map(
        lambda idx: list(model.parameters())[idx], bucket_indices_backward
    )
    for bucket in params_in_buckets:
        size_bytes = sum(p.numel() * p.element_size() for p in bucket)
        size_mb = round(size_bytes / 1024 / 1024, 3)
        bucket_sizes.append(size_mb)

    return bucket_sizes


def _convert_ids_int_string(slices: List[Any]) -> None:
    """
    Used by get_perfetto_object to convert ids to str
    Inputs: List of slices
    Outputs: None
    """
    for slice in slices:
        if "id" in slice:
            slice["id"] = str(slice["id"])


def _convert_negative_tids_to_positive(slices: List[Any]) -> None:
    """
    Used by get_perfetto_object to change ids from neg to pos
    Inputs: List of slices
    Outputs: None
    """
    for slice in slices:
        if "tid" in slice and isinstance(slice["tid"], int):
            slice["tid"] = abs(slice["tid"])


def _remove_args(slices: List[Any]) -> None:
    """
    Used by get_perfetto_object to remove unused args
    Inputs: List of slices
    Outputs: None
    """
    [slice.pop("args", None) for slice in slices]


def get_perfetto_object(filepath: str) -> TraceProcessor:
    """
    Input: Pytorch profiler trace
    Output: Handler to run SQL-like queries on trace. 
    """
    with open(filepath, "rb") as f:
        raw_slices = orjson.loads(f.read())

    if isinstance(raw_slices, dict) and "traceEvents" in raw_slices:
        # Perfetto doesn't want this format produced by PyTorch
        raw_slices = raw_slices.pop("traceEvents", None)
    # Convert IDs from int to string. Without this perfetto fails to JSON load trace with IDs stored as integers.
    _convert_ids_int_string(raw_slices)
    # Convert negative 'tid' values to positive. Without this perfetto combines together the slices with different tids into one track
    _convert_negative_tids_to_positive(raw_slices)
    _remove_args(raw_slices)  # For speedup

    slices_bytes = orjson.dumps(raw_slices)
    slices_bytes = io.BytesIO(slices_bytes)  # 4x speedup using orjson

    try:
        tp = TraceProcessor(slices_bytes)
    except ConnectionResetError as e:
        # This happens sometimes so retry once
        tp = TraceProcessor(slices_bytes)

    interesting_fields = "SELECT ts, dur, track_id, category, name, depth, cat, slice_id, id, arg_set_id FROM slice"

    def query_dict(query):
        query = query.lower().replace(
            "select * from slice", interesting_fields
        )  # gives a 15% speedup
        try:
            query_iterator = tp.query(query)
        except Exception as e:
            print("[ERROR] Unable to run query: %s" % query)
            print(traceback.format_exc())
            print("\n\n")
            raise e
        return [item.__dict__ for item in query_iterator]

    tp.query_dict = query_dict

    return tp


def read_gpu_slice(tp: TraceProcessor, cpu_input_slice: dict) -> dict:
    """
    Input: Perfetto handler and CPU slice
    Output: GPU kernel slice
    """
    cuda_slice = None
    slice_id_origin = cpu_input_slice["slice_id"]
    slice_id_destination = tp.query_dict(
        f"""select * from flow where slice_out={slice_id_origin}"""
    )

    if slice_id_destination:
        cuda_slice = tp.query_dict(
            f"select * from slice where slice_id={slice_id_destination[0]['slice_in']}"
        )[0]

    return cuda_slice


def get_first_last_step(filepath: str) -> Tuple[int, int]:
    """
    Input: Pytorch profiler trace
    Output: First and last step of the trace.
    """
    tp = get_perfetto_object(filepath)
    steps_arr = tp.query_dict(f"select * from slices where name like '%ProfilerStep#%'")
    first_step = int(steps_arr[0]["name"].split("#")[1])
    last_step = int(steps_arr[-1]["name"].split("#")[1])

    return first_step, last_step


def get_ddp_forward_backward_times(filepath, step) -> Tuple[float, List[float]]:
    """
    Inputs: Pytorch profiler trace and step number
    Outputs: Forward runtime and list of ddp-bucket computation times
    """
    tp = get_perfetto_object(filepath)
    profiler_step = tp.query_dict(
        f"select * from slices where name like '%ProfilerStep#{step}%'"
    )[0]
    start_step = profiler_step["ts"]
    end_step = profiler_step["ts"] + profiler_step["dur"]
    forward_track = profiler_step["track_id"]
    backward_track = tp.query_dict(
        f"""
                                    SELECT track_id from slices
                                    WHERE name like '%autograd::%'
                                    AND ts> {start_step} and ts < {end_step}
                                    """
    )[0]["track_id"]

    ## ==================== GET TOTAL BACKWARD TIME =================== ##

    backward_slices = tp.query_dict(
        f"""
                                        select * from slices where track_id={backward_track}
                                        AND ts > {start_step} AND ts < {end_step}
                                        """
    )
    
    start_backward, end_backward = (
        backward_slices[0]["ts"],
        backward_slices[-1]["ts"] + backward_slices[-1]["dur"],
    )

    ## ==================== GET DEVICE FORWARD TIME =================== ##
    forward_slices = tp.query_dict(
        f"""
                                    select * from slices
                                    where ts > {start_step} AND ts < {start_backward}
                                    AND track_id={forward_track}
                                    """
    )
    start_forward, end_forward = (
        forward_slices[0]["ts"],
        forward_slices[-1]["ts"] + forward_slices[-1]["dur"],
    )

    forward_cuda_calls = tp.query_dict(
        f"""
                                        select * from slices where ts > {start_forward} AND ts < {end_forward}
                                        AND name like '%cudaLaunchKernel%'
                                        """
    )

    forward_first_cuda_call = forward_cuda_calls[0]
    forward_last_cuda_call = forward_cuda_calls[-1]
    forward_device_start = 0
    forward_device_ends = 0
    if forward_first_cuda_call:
        cuda_slice_start = read_gpu_slice(tp, forward_first_cuda_call)
        forward_device_start = cuda_slice_start["ts"] if cuda_slice_start else 0
    if forward_last_cuda_call:
        cuda_slice_end = read_gpu_slice(tp, forward_last_cuda_call)
        forward_device_ends = (
            cuda_slice_end["ts"] + cuda_slice_end["dur"] if cuda_slice_end else 0
        )

    forward_start_ts = (
        forward_device_start if forward_device_start != 0 else start_forward
    )
    forward_end_ts = forward_device_ends if forward_device_ends != 0 else end_forward

    ## ==================== GET BUCKET TIMES =================== ##
    find_c10_all_reduce_calls = f"""
                                SELECT * from slice main
                                WHERE main.track_id={backward_track}
                                AND main.ts > {start_backward} AND main.ts < {end_backward}
                                AND main.name like '%autograd::engine::evaluate_function: torch::autograd::AccumulateGrad%'
                                AND 'c10d::allreduce_' IN (SELECT submain.name FROM slice submain WHERE submain.ts > main.ts AND submain.ts < main.ts + main.dur AND submain.track_id={backward_track} )    
                                """
    all_reduce_calls = tp.query_dict(find_c10_all_reduce_calls)
    prev_ts = start_step
    bucket_comp_times = []
    for idx, item in enumerate(all_reduce_calls):
        # ================== COMPUTATION ======================================#
        end_ts = item["ts"]
        slices_in_bucket = tp.query_dict(
            f"""
                                         select * from slice
                                         where track_id={backward_track}
                                         and ts > {prev_ts} and ts < {end_ts}                
                                        """
        )
        start_cpu_time = slices_in_bucket[0]["ts"]
        end_cpu_time = slices_in_bucket[-1]["ts"] + slices_in_bucket[-1]["dur"]

        device_ts_start = 0
        device_ts_end = 0

        cuda_launch_list = tp.query_dict(
            f"""
                                        select * from slice
                                        where ts > {start_cpu_time} and ts < {end_cpu_time} and track_id={backward_track}
                                        and name like '%cudaLaunchKernel%'
                                        """
        )
        if cuda_launch_list:
            cuda_slice_start = read_gpu_slice(tp, cuda_launch_list[0])
            device_ts_start = cuda_slice_start["ts"] if cuda_slice_start else 0

            cuda_slice_end = read_gpu_slice(tp, cuda_launch_list[-1])
            device_ts_end = (
                cuda_slice_end["ts"] + cuda_slice_end["dur"] if cuda_slice_end else 0
            )

        netCompTime = max(0, device_ts_end - device_ts_start)

        bucket_comp_times.append(round(netCompTime * 1e-6, 3))

        prev_ts = item["ts"] + item["dur"]

    forward_time = round((forward_end_ts - forward_start_ts) * 1e-6, 3)

    return forward_time, bucket_comp_times
